//=========================================================================
//  SEEDTOOL.CC - part of
//
//                  OMNeT++/OMNEST
//           Discrete System Simulation in C++
//
//   seedtool utility
//
//   To compile under Unix: gcc seedtool -o seedtool
//
//=========================================================================

/*--------------------------------------------------------------*
  Copyright (C) 1992-2005 Andras Varga and Gyorgy Pongor,

  This file is distributed WITHOUT ANY WARRANTY. See the file
  `license' for details on this and other legal matters.
*--------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>

#define INTRAND_MAX  0x7ffffffeL  /* = 2**31-2 */

#define TABLESIZE   4096
struct hashelem {long index; long seed;};
extern struct hashelem hashtable[TABLESIZE];

/*------- long int pseudorandom number generator -------------
*    Range:          1 ... 2**31-2
*    Period length:  2**31-2
*    Global variable used:   long int theSeed : seed
*    Method:  x=(x * 7**5) mod (2**31-1)
*    To check:  if  x[0]=1  then  x[10000]=1,043,618,065
*    Required hardware:  exactly 32-bit integer aritmetics
*    Source:  Raj Jain: The Art of Computer Systems Performance Analysis
*                (John Wiley & Sons, 1991)   Pages 441-444, 455
*---------------------------------------------------------*/

long nextrand(long seed)
{
     const long int a=16807, q=127773, r=2836;
     seed=a*(seed%q) - r*(seed/q);
     if (seed<=0) seed+=INTRAND_MAX+1;
     return seed;
}

int testrand()
{
     long seed = 1;
     for(int i=0; i<10000; i++) seed=nextrand(seed);
     int good = (seed==1043618065L);
     return good;
}

void filltable()
{
    const long cyclelength = 0x7FFFFFFE;
    const long gap = cyclelength/TABLESIZE;

    long i, seed, nextstop, k;

    // init the table
    for(i=0; i<TABLESIZE; i++)
        hashtable[i].index = hashtable[i].seed = -1;

    // fill out the table
    i=0; seed=1; nextstop=0; k=0;
    do {
        if (i>=nextstop)
        {
            // generate hash value out of seed
            int hashvalue = seed%TABLESIZE;

            // insert it into the table if its slot is unused
            if (hashtable[hashvalue].index==-1)
            {
                hashtable[hashvalue].index = i;
                hashtable[hashvalue].seed = seed;
                nextstop += gap;

                fprintf(stderr,"\r%ld",++k);
                fflush(stderr);
            }
        }

        // next seed value
        seed = nextrand(seed);
        i++;
    } while (seed!=1);
    printf("\n");
}

void printtable()
{
    printf("struct hashelem  hashtable[TABLESIZE] = {\n");
    for(int i=0;i<TABLESIZE;i++)
    {
        printf( "{%ld,%ld},",
                hashtable[i].index,
                hashtable[i].seed
              );
        if (i%4==3) printf("\n");
    }
    printf("\n};");
}

long indexofseed(long seed)
{
    // check seed passed
    if (seed<=0 || seed>INTRAND_MAX)
    {
         fprintf(stderr,"Invalid seed %ld\n",seed);
         exit(1);
    }

    // find closest match in the table
    long i=0;
    for(;;)
    {
        // generate hash value out of seed
        int hashvalue = seed%TABLESIZE;

        // is it in the table?
        if (hashtable[hashvalue].seed==seed)
            return hashtable[hashvalue].index - i;

        // next seed value
        seed = nextrand(seed);
        i++;
    }
}

long seedatindex(long index)
{
    // find closest index in hashtable
    int closest = -1;
    long i;
    for (i=0; i<TABLESIZE; i++)
        if (hashtable[i].index<=index && (closest<0 || hashtable[i].index>hashtable[closest].index))
            closest = i;

    // cycle to desired index "on foot"
    long seed = hashtable[closest].seed;
    for(i = hashtable[closest].index; i<index; i++)
        seed = nextrand(seed);
    return seed;
}

int main(int argc, char *argv[])
{
    if (argc==1)
    {
         fprintf(stderr,
               "seedtool - part of OMNeT++/OMNEST, (C) 1992-2005 Andras Varga\n"
               "See the license for distribution terms and warranty disclaimer.\n"
               "\n"
               "Generates seeds for the LCG32 random number generator. This RNG has a\n"
               "period length of 2^31-2, which makes about 2.147 billion random numbers.\n"
               "Note that Mersenne Twister is also available in OMNeT++, which has a\n"
               "practically infinite period length (2^19937).\n\n"
                "Usage:\n"
                "  seedtool i seed         - index of 'seed' in cycle\n"
                "  seedtool s index        - seed at index 'index' in cycle\n"
                "  seedtool d seed1 seed2  - distance of 'seed1' and 'seed2' in cycle\n"
                "  seedtool g seed0 dist   - generate seed 'dist' away from 'seed0'\n"
                "  seedtool g seed0 dist n - generate 'n' seeds 'dist' apart, starting at 'seed0'\n"
                "  seedtool t              - generate hashtable\n"
                "  seedtool p              - print hashtable\n"
                );
         exit(0);
    }

    if (!testrand())
    {
         fprintf(stderr,"Random number generator does not work correctly!\n");
         exit(1);
    }

    long arg[5];
    for (int i=2; i<5; i++)
        arg[i] = argc>i ? atol(argv[i]) : 0;

    switch (argv[1][0])
    {
      case 'i':
        printf("%ld\n", indexofseed( arg[2] ));
        break;
      case 's':
        printf("%ld\n", seedatindex( arg[2] ));
        break;
      case 'd':
        printf("%ld\n", indexofseed( arg[3] ) - indexofseed( arg[2] ));
        break;
      case 'g':
        {
            long index = indexofseed( arg[2] ),
                 dist = arg[3],
                 count = arg[4];
            if (count==0) count=1;

            /*
             * 20020806, N. Reijers
             * This runs into trouble if too many seeds, too far apart are requested.
             * There are 2^31 seeds, but when we keep adding dist to index, it will
             * eventually exceed 2^31.
             * Because index is of type long, it doesn't loop until 2^32, so we get
             * periods of 2^31/dist seeds followed by as many zeros.
             * The fix applied here is the line index&=0x7fffffffL; which makes it loop.
             * (effectively it does a index%=2^31;)
             * The result is that a continuous stream of seeds is generated although,
             * depending on dist, the period will be somewhere between 2^31/dist and
             * 2^31.
             */
            for (; count>0; count--)
            {
                 index += dist;
                 index &= 0x7fffffffL; /* Fix. */
                 printf("%ld\n", seedatindex( index ));
            }
        }
        break;
      case 't':
        filltable();
        printtable();
        break;
      case 'p':
        printtable();
        break;
      default:
        fprintf(stderr,"Invalid switch %s\n",argv[1]);
        return 1;
    }
    return 0;
}

//==========================================================================
// The following table was automatically generated with 'seedtool t > file'

struct hashelem hashtable[TABLESIZE] = {
{225443410,448950272},{0,1},{881326447,1096794114},{382205223,350486531},{1603793934,1682608132},
{1465382165,1332060165},{328203662,1203453958},{205520504,682684423},{665844490,662712328},
{201326209,1944424457},{1867510295,1134006282},{888142178,2036408331},{524811287,1442787340},
{190316182,170766349},{1582298168,221458446},{175111858,1634951183},{1707078472,1131225104},
{870840707,1356865553},{91225938,713355282},{1952969076,832454675},{1400370577,709382164},
{1819800181,987041813},{235929150,1975304214},{111673131,1846829079},{1023932513,69812248},
{1766847199,1719595033},{946862322,761995290},{1387787689,236412955},{461372560,2035470364},
{352320864,1209155613},{6291444,1616572446},{763886159,861655071},{883423595,1553391648},
{1282930289,2021711905},{1804595861,2011299874},{1824518770,1297158179},{1645212612,1569869860},
{1347941877,1370636325},{397933833,568291366},{562559951,378998823},{525335574,103219240},
{1903686101,584294441},{1525675178,1278636074},{1931473310,1154060331},{1422390632,2018992172},
{2068836517,4141101},{689961692,1996484654},{768080455,2000973871},{745011827,1179041840},
{855636386,2039369777},{2063593647,1644945458},{39845812,1226514483},{1821897341,1541886004},
{1299707473,959442997},{769653317,990683190},{715651756,932954167},{1808265881,839446584},
{16777184,583516217},{1805120145,1879564346},{1247278773,1752514619},{849869227,695664700},
{1828188776,1174114365},{1983377730,2111402046},{2096099538,1623355455},{1870656024,542462016},
{1296561755,412872769},{1943007639,1629278274},{1339553286,758440003},{1335883277,767418436},
{331349384,1461260357},{1696068448,803938374},{1543500933,823246919},{1618473969,1463713864},
{1555035242,393289801},{1734865688,351338570},{329776524,1934868555},{1303377484,585986124},
{1447032120,1571516493},{1386214828,1811361870},{1922560429,1798586447},{885520744,1159557200},
{331873671,1853395025},{1265628819,1334165586},{847247793,426844243},{80215911,953606228},
{310377904,234786901},{1576006722,841936982},{1366816209,1546641495},{1666708375,1373540440},
{1409283461,1349828697},{1936191895,1183199322},{155713239,1965199451},{298843590,599105628},
{1709175623,869097565},{1752691443,1780707422},{102760252,1571070047},{1931997608,1597710432},
{1085798377,245035105},{217579105,17199202},{1526723746,1141731427},{1656222636,348524644},
{1393554850,561471589},{182451877,997285990},{675281656,538529895},{212336235,1676873832},
{578288561,333746281},{1365767637,1868222570},{1594356771,606978155},{130023176,161980524},
{989853856,1840164973},{139984629,1245012078},{969406663,2017685615},{1521480879,1103036528},
{1287124591,734310513},{1707602760,354345074},{1970794834,1585688691},{1807217293,1341907060},
{1809314441,1529860213},{372768057,140619894},{447216811,1988796535},{734001800,1621143672},
{1085274091,954638457},{19398619,2127011962},{718797477,401801339},{515374121,86974588},
{1129314202,1538003069},{2021650683,760225918},{229637706,1098535039},{2028990730,1787859072},
{1455944999,1352798337},{1204811526,1900073090},{1047525426,1663127683},{217054818,965865604},
{997718161,812322949},{2038952172,751128710},{682097389,1563746439},{430963914,215142536},
{1703932755,19280009},{339737976,1581576330},{591395736,946602123},{1216870127,821276812},
{1094711256,685330573},{2138566769,2058440846},{1417147762,588292239},{938998017,1416867984},
{743963254,113270929},{1647834043,461963410},{328727949,438349971},{294649294,1662521492},
{1018165355,2083369109},{1724904237,1358074006},{139460342,1904025751},{1266677392,1868091544},
{664271629,765603993},{497024076,235450522},{935328008,486404251},{18874332,757579932},
{1138227080,2035470493},{1355281897,541892766},{1172305732,1597907103},{407370999,52568224},
{241172020,1684254881},{1274541700,823693474},{79691624,244301987},{1095759830,1912180900},
{1889006073,1480372389},{425721045,996950182},{1376253375,1175937191},{2120741100,1690255528},
{1974989129,551452841},{180354728,1989922986},{1473246470,1428873387},{1118828458,1247629484},
{1307571778,696344749},{447741098,613052590},{673184508,684441775},{2000679199,1558978736},
{1112012727,1023471793},{1665659806,1986539698},{1286600299,434897075},{2069360798,948887732},
{1757410033,1392574645},{622328670,269410486},{860879254,2020102327},{959445211,541262008},
{1394079138,564428985},{1981804885,1396134074},{236453437,1300271291},{1715991352,439771324},
{1762128613,2129072317},{1558705255,415916222},{115343140,2141974719},{1366291922,1618170048},
{573045693,504606913},{587725727,1391653058},{943192313,905875651},{1986523452,1894858948},
{1679291261,927056069},{61865866,280236230},{1792537254,463552711},{34078655,1483522248},
{121110297,701550793},{963115219,412168394},{584055718,1981071563},{1300756047,652255436},
{1933046181,1526038733},{1486353649,522199246},{1804071570,816099535},{679475952,1220571344},
{915405102,1087430865},{1868034585,319443154},{1973416278,1264013523},{948959470,1779298516},
{455605403,1213071573},{179830441,682369238},{90177364,772923607},{712506033,1402732760},
{1100478414,22339801},{1728049961,1190383834},{952629479,1410621659},{879753586,1031110876},
{150994656,1148367069},{782236204,1348813022},{1733817113,1583128799},{1568666705,892080352},
{1371534792,280383713},{1468527887,303300834},{1928851884,1881854179},{574094266,567697636},
{2107633749,1513136357},{1078982649,1399558374},{1677718400,755478759},{726137496,1033433320},
{665320205,199610601},{1624765417,603807978},{276823536,279015659},{446692524,1793765612},
{1700787051,746139885},{278920684,171671790},{1177024317,657981679},{283114980,736784624},
{661650194,108441841},{136314620,893821170},{1812984451,1636835571},{1416099187,1533169908},
{1188034342,1813913845},{1404040587,1128841462},{1577579591,703070455},{1096284117,634507512},
{1244657338,2053030137},{1506276554,1931268346},{1848635962,1411055867},{1007679614,292278524},
{83885920,119038205},{507509816,183132414},{1454372140,682860799},{846199219,319418624},
{241696307,854950145},{1962406258,157438210},{1222637284,1719038211},{1120401320,1970266372},
{249560612,1496523013},{1580725307,1354875142},{1360000485,1139630343},{1832907352,1183895816},
{2111828060,1267101961},{1906307533,1463787786},{265813509,1873330443},{1373107653,1944666380},
{1596978203,657486093},{196607625,609943822},{714603181,1783611663},{1297086042,1327038736},
{419429600,773591313},{1880617494,800997650},{1363146201,417493267},{1793585831,364245268},
{1406662023,705237269},{819460581,305926422},{281017832,960590103},{752351849,747389208},
{1717039927,386728217},{1847063110,237388058},{1943531926,989171995},{1109915579,386056476},
{940046591,1309557021},{1863316017,737956126},{1274017410,913649951},{1169684297,975909152},
{1709699919,1252188449},{1641018315,1086968098},{171966136,768741667},{1214772981,388469028},
{1666184087,1371423013},{418905313,958353702},{838334913,674476327},{1402467726,1383403816},
{2142237181,133194025},{316145061,1622049066},{1858597421,1016860971},{980940977,1847382316},
{1410856317,1945039149},{605551485,152727854},{208141939,772092207},{1491596516,394846512},
{206569078,1320091953},{651164455,1115742514},{1782051522,902844723},{904395076,1992180020},
{1639969736,483295541},{1609561090,1210581302},{93323086,1661600055},{1170732871,764473656},
{992999578,1543221561},{1268250254,1752793402},{2061496493,756404539},{1070069767,1535365436},
{908065084,2119835965},{105905974,447004990},{203947643,745017663},{1902637525,2035323200},
{1470625036,456720705},{1293940317,966865218},{2031087893,515277123},{1580201020,1433211204},
{761789011,239944005},{589822876,1407811910},{2124935221,765800775},{1458566434,1583567176},
{322436506,577896777},{2070409393,1667531082},{92798799,1305207115},{1402992013,695619916},
{26214350,1239789901},{1662514078,865120590},{1187510055,1494831439},{923269407,1179930960},
{1653601199,1462825297},{355466589,833524050},{49282978,1776021843},{1201141517,1699782996},
{1352136174,163905877},{1056438307,800264534},{727186069,1168757079},{486538337,2032382296},
{2019029263,2012655961},{721418913,259957082},{375913779,1139999067},{1494217950,1922802012},
{287833563,1118753117},{1892676077,1334284638},{1265104531,1826054495},{1453323565,2119528800},
{1857024561,19644769},{567802823,410636642},{2024272129,246387043},{207617652,1821061476},
{1905258961,877080933},{1390409127,537301350},{815266285,1190539623},{829422034,754233704},
{1775235782,634716521},{1747448577,412754282},{365428039,688894315},{610794355,1908306284},
{1353709035,1776075117},{468712581,520593774},{1533539475,1928278383},{1379923386,1891606896},
{123207445,1588994417},{637532992,1275359602},{1643639745,617234803},{342883698,1532207476},
{1595405341,421708149},{431488201,952549750},{2073555124,2056384887},{301989312,194961784},
{2118119516,1162678649},{1447556412,1647120762},{699398860,1790841211},{1881141757,1413079420},
{1403516300,222781821},{1739584266,659394942},{608697207,152785279},{229113419,963457408},
{26738637,89227649},{1142945660,251556226},{1715467064,1949761923},{737147522,1698099588},
{2075128010,386232709},{910686520,219640198},{131596037,1334854023},{325057940,342524296},
{1126168476,1507574153},{1321727527,1598329226},{1482683639,939680139},{1004009606,853021068},
{414186730,1340805517},{1247803062,1294057870},{2041573604,1167884687},{1388836263,1617457552},
{267386370,1108701585},{78643050,522199442},{220200540,1523306899},{1712321342,1627132308},
{654310176,710447509},{1528296607,1003721110},{198180486,1875599767},{9961453,1408156056},
{1476916483,250286489},{2039476456,171946394},{389545241,2082066843},{187694746,436801948},
{775420473,1773719965},{1814557318,871215518},{285212128,389554591},{1957163381,806281632},
{1415574900,2020360609},{615512938,2143486370},{697825997,2013491619},{197131912,764506532},
{809499129,1156997541},{1009776763,387789222},{137887482,886022567},{1833955927,1835553192},
{494926929,914334121},{581958570,80150954},{2036855003,1973367211},{1634726866,1320157612},
{2029515000,334365101},{1211102973,632848814},{1769468630,336855471},{1822945912,880603568},
{911210806,2057609649},{406322425,640250290},{606600059,627925427},{1956639087,491418036},
{1292367458,1340854709},{1571288141,723870134},{374865205,2048905655},{874510716,283570616},
{68157310,1996652985},{1024981085,76878266},{1516762292,1209823675},{338165115,1762640316},
{857733532,108880317},{305135034,1893245374},{1873801745,1216299455},{1213724406,1542459840},
{1585968177,1817665985},{859830680,483254722},{645921584,1849590211},{2087710940,439919044},
{937949443,1703489989},{1238890188,474550726},{1810363016,241775047},{1283978864,825377224},
{715127468,2068455881},{1153955688,1010479562},{518519843,1745736139},{1263007383,582656460},
{1687679854,899744205},{1223161574,1844257230},{18350045,1142485455},{1728574241,1136140752},
{2052059360,540918225},{1022883938,1662480850},{458751125,1359208915},{1518859439,524255700},
{825752025,1115988437},{1598026778,242475478},{511704113,114799063},{1667756949,1028055512},
{1787818670,480182745},{596114319,704717274},{561511377,184922587},{1310717500,700412380},
{1801974425,462004701},{2095050976,941228510},{1224734432,303423967},{1839198801,2077503968},
{491781206,634847713},{527957009,1907565026},{1043855417,2132156899},{423623896,1551839716},
{275774963,448164325},{1308620356,955159014},{841480635,1229967847},{1571812426,1636078056},
{906492225,1385718249},{1480062201,1401322},{1277163133,1694519787},{919075111,870998508},
{238026299,1025671661},{644348723,440254958},{243269169,1943200239},{1617425400,2062193136},
{238550585,758108657},{896530771,103227890},{55574422,1625440755},{532675592,579711476},
{1146615674,1145602549},{753924706,528175606},{965212367,359739895},{1317008946,950125048},
{1057486880,836334073},{932182287,1248637434},{1798828697,1680572923},{1206384388,1700594172},
{74448754,545853949},{1632105433,338485758},{1885336055,336675327},{1917841854,1498530304},
{1409807743,1217208833},{721943199,756646402},{249036325,2083340803},{1048574001,1124229636},
{670563073,2125804037},{1302328909,2088550918},{1334310424,1383408135},{1439167819,230781448},
{296746442,882360841},{823654877,439783946},{526908437,1398530571},{1003485318,1928045068},
{772274751,615481869},{459799699,1760997902},{758643289,1394549263},{1162344279,1931162128},
{1441264969,1144525329},{1199044369,1040359954},{1551365239,610927123},{1194850077,1686118932},
{370670909,1378673173},{244842029,1898611222},{1652552628,596382231},{519044130,1669251608},
{973600959,172319257},{1461187869,310166042},{1838150239,745194011},{209714800,1668895260},
{2129129556,1401029149},{1818227318,1121772062},{1041758271,1397797407},{290979286,617054752},
{1829761642,423080481},{280493546,284234274},{399506696,1403339299},{330300810,146104868},
{384302371,795451941},{1270347401,1707000358},{2035806426,2006471207},{1597502494,1930011176},
{1548743799,1036304937},{1454896426,59396650},{643824436,1848476203},{1246754486,769479212},
{1394603420,1071018541},{99614530,1412665902},{1017641069,207086127},{89128790,1882346032},
{1159722844,803848753},{970979524,418099762},{887617891,1596236339},{799537676,1715384884},
{141557490,1673241141},{988805287,2103448118},{2092429437,1342468663},{148373221,1540575800},
{1323824675,1215291961},{843053496,621048378},{1014495346,1333916219},{1305474631,1740681788},
{71303032,994071101},{1005582467,389472830},{242744881,1490403903},{2137518328,2050056768},
{415759591,670868033},{1429730652,451547714},{1135081355,278295107},{1650979765,262685252},
{1972367698,1372213829},{566229960,1470370374},{289930711,993755719},{197656199,1338212936},
{476052596,846869065},{1046476852,1285370442},{614988651,2065404491},{668990213,980386380},
{176160432,186901069},{730856078,413872718},{538442749,921399887},{1307047495,828600912},
{1525150885,1242006097},{119537436,859427410},{405798138,686436947},{266337796,575005268},
{16252897,2048385621},{1395127707,1810309718},{180879015,1220502103},{1961881971,1074836056},
{253230621,446206553},{1685582710,758506074},{975698107,1390232155},{660601620,757023324},
{264764935,1194664541},{1912074696,877494878},{2050486462,267686495},{749206126,1731203680},
{1489499368,2133787233},{427293905,835822178},{432536775,777347683},{1630008289,116683364},
{380108075,1325478501},{1786770101,432820838},{1295513178,1210339943},{1795158689,1660019304},
{1959784827,1973809769},{891812187,1400492650},{1497363673,487453291},{1602221072,741380716},
{1562899550,20099693},{2133848196,542495342},{950532331,2004058735},{632814410,264606320},
{1814033020,616407665},{1141372799,1422578290},{575667126,1549939315},{1325397538,1521242740},
{1703408463,2098369141},{555744220,1678365302},{1894248945,1701409399},{1819275896,762831480},
{555219933,947712633},{914356528,1011380858},{2003300629,674849403},{61341579,1951003260},
{252182047,1736905341},{1320678954,642978430},{1805644442,1318122111},{314047913,1627824768},
{1432876373,1184318081},{243793455,1021645442},{1967649114,136426115},{1180170037,1880175236},
{473431161,1049055877},{275250675,1403032198},{1344796158,1200743047},{33554368,375841416},
{1277687422,2133287561},{1181218615,1669272202},{770701890,1790288523},{492305494,622363276},
{1453847853,205054605},{1424487783,1797800590},{725088921,2083693199},{1713894208,73597584},
{1022359652,1775657617},{1965552044,414724754},{509082677,744727187},{1158149984,1722872468},
{227540558,945074837},{1863840293,227414678},{889190752,336118423},{695728852,333742744},
{1813508735,1272316569},{744487540,266822298},{1190131490,226472603},{2135421024,1556632220},
{2071457952,498422429},{1108342718,448852638},{616561512,630108831},{142081777,1722913440},
{1020262502,2123236001},{1851781690,1970442914},{1118304172,1354695331},{1833431639,270598820},
{836237766,141419173},{685767397,360829606},{1415050615,963326631},{474479735,991765160},
{1503655117,1784926889},{1372583368,999850666},{2040525032,330527403},{1946153352,343999148},
{1428682081,1057313453},{1751642872,2098676398},{245890605,772182703},{1596453915,1635377840},
{1331688982,1854411441},{1576531011,754078386},{610270068,2051871411},{1758458598,2121441972},
{2012737793,1593565877},{2130178133,456172214},{178781867,1389458103},{1724379944,2123408056},
{322960792,3637945},{1713369918,2123006650},{362282317,1046753979},{1438643528,462066364},
{1749021436,397714109},{1477440766,750764734},{1604842508,720646847},{1226831580,1863037632},
{115867427,1520370369},{1615328247,1803371202},{98565956,889959107},{133693185,729789124},
{378535214,341709509},{1374156227,1849303750},{464518282,881156807},{633862983,1449153224},
{929036565,1801122505},{462421134,245068490},{481295466,12493515},{1536685198,703083212},
{2058875053,1705562829},{1166538575,746873550},{760740437,1541485263},{47710117,1210413776},
{2139615272,1541599953},{420478176,1635365586},{659553046,81654483},{2033184993,373154516},
{1646785469,1306583765},{1544025215,277115606},{1001912457,1705755351},{292552146,493781720},
{2131750995,502682329},{202374782,328319706},{1237317320,1926025947},{158858961,1640567516},
{451935394,1426592477},{1246230199,478667486},{993523865,1403101919},{902822215,471765728},
{2020077834,977625825},{746584688,660927202},{1215821554,1785082595},{417332453,1374667492},
{11534314,136225509},{1330116119,2143507174},{1627386854,245875431},{1549268085,995078888},
{1490547946,1729934057},{403700990,224875242},{1799352984,984007403},{1090516960,13058796},
{183500450,147034861},{1507325127,400667374},{287309276,484479727},{1097856980,1401193200},
{2123362420,702165745},{1468003601,608064242},{882899308,426533619},{732428940,2041852660},
{1519908014,1338516213},{754973280,2073772790},{623901531,48321271},{1133508495,1443205880},
{1537209488,2007913209},{1928327596,1722561274},{222821975,1604133627},{748681839,1199915772},
{1485829371,1376244477},{3145722,728359678},{1237841608,1022632703},{1751118580,1035412224},
{2621435,150205185},{1324873249,242094850},{494402641,265822979},{2065166517,746640132},
{24117202,1451713285},{722991773,56218374},{620755809,2039157511},{830994895,71783176},
{1976037711,548582153},{933755147,1184768778},{702544581,2026271499},{1208481536,110936844},
{1365243349,2120033037},{457702551,1423966990},{477625457,1651860239},{350748003,503603984},
{739768957,1943507729},{1678766975,650330898},{1450702129,2109047571},{853539236,496935700},
{1012922488,1501635349},{247987751,40657686},{905443654,607716119},{258473492,1431008024},
{34602942,1112441625},{703593154,1715528474},{1144518522,1605587739},{706214590,2129408796},
{1211627259,2014663453},{1240463044,2000708382},{999291022,820794143},{1659892660,1150464800},
{112721705,1499480865},{1652028340,16724770},{1493169376,253748003},{268959232,1222226724},
{858257821,1051747109},{682621674,1619288870},{1343747581,498746151},{1313863223,1123742504},
{1340601862,1964380969},{1210578685,917152554},{1947201919,1053733675},{1354757608,2134696748},
{822606303,1491313453},{437779645,1793241902},{1839723086,1313755951},{2134372909,1029063472},
{613940077,1257632561},{101711678,442782514},{1000863883,1019794227},{128974603,164143924},
{273153527,317731637},{911735093,1906778934},{1220540136,1374077751},{867170699,829567800},
{1109391292,433591097},{2117595230,1660711738},{930609425,529793851},{598211467,1952461628},
{1390933411,384021309},{1719137075,1225974590},{4194296,1550086975},{1419244912,396882752},
{1547170938,1984639809},{1474819344,353272642},{1215297268,1813091139},{731380365,706302788},
{1589638185,1752580933},{824179164,740062022},{1063254036,1963918151},{2038427863,1085518664},
{369622336,1585439561},{30932933,1651938122},{563084238,1201251147},{461896848,160904012},
{1002436744,153793357},{907540799,302342990},{735050374,973132623},{585628580,1794274128},
{125304593,14717777},{728758932,244613970},{1991766316,1531274067},{1758982894,1920463700},
{1362097628,1148158805},{1032321103,1870893910},{833092043,1714385751},{751303271,1446937432},
{519568417,912171865},{1411380608,1790341978},{1321203240,191075163},{543685619,1471460188},
{626522965,383157085},{1126692763,192324446},{612891503,58975071},{281542119,1480500064},
{511179828,1327940449},{1927803308,1583481698},{622852956,911848291},{422575322,1904640868},
{1677194114,689271653},{1431827797,540447590},{1958736245,1785766759},{1130362772,1801823080},
{2031612127,2135397225},{1081079794,2052182890},{1101002701,1541321579},{926415130,1888969580},
{1466430741,718402413},{1656746920,1290359662},{1637872594,1452168047},{648018733,1840223088},
{1729622818,1526322033},{1383069107,560468850},{348126568,404091763},{1731195688,1444303732},
{1732768536,1340859253},{392690963,408101750},{1585443893,1796735863},{7340018,1385505656},
{1093662682,1171805049},{1298658901,1362133882},{92274512,708420475},{248512038,1348117372},
{602930050,1020441469},{588774301,1552450430},{973076674,462443391},{1676669828,1450644352},
{2145383741,255046529},{2109206636,2016093058},{1689252719,1288926083},{1151334252,346723204},
{132120324,117556101},{1089468387,1865302918},{1552938094,1174578055},{1335358989,1753465736},
{1952444789,906064777},{1622668267,401580938},{2051535038,1581855627},{1023408224,959325068},
{1606415368,1406149517},{320339357,19461006},{1357903330,992605071},{1083176942,525861776},
{1509946561,1286706065},{1399322003,29942674},{1010825336,146568083},{2130702639,2077799316},
{1714418503,44536725},{493878354,1018192790},{1185412907,1037898647},{416283879,253301656},
{372243770,159413145},{1920463281,543626138},{1374680514,88011675},{163577544,1365582748},
{960493787,1340384157},{889715039,1556034462},{563608526,684975007},{995096728,833282976},
{1312290362,495899553},{975173821,460362658},{1773138636,1637700515},{890239327,142980004},
{1897394653,1137750949},{1131935638,1269851046},{1520956594,1557476263},{1385166254,1914418088},
{1106245571,1532056489},{704117444,655098794},{1391981985,1084531627},{535297027,796853164},
{777517621,1897907117},{267910657,1582953390},{1430254936,694817711},{395312398,1571926960},
{1970270560,91730865},{836762052,897418162},{687864544,106570675},{2023747862,637625268},
{44040108,1238676405},{2096623786,730346422},{564132813,531325879},{1451750711,397079480},
{1676145539,1174807481},{1174402881,69665722},{589298588,1636438971},{1940386193,1669129148},
{155188952,292090813},{1317533231,1103176638},{1698165596,157828031},{2032660765,1649947584},
{839383487,1865630657},{1929900447,1718506434},{1066924045,37135299},{531627018,1533617092},
{1996484905,2103006149},{838859202,1795085254},{1950347652,261653447},{2098720877,1293038536},
{851966375,1555624905},{1002961032,2125689802},{1417672049,1730634699},{850393514,492270540},
{86507355,600740813},{912783667,1643160526},{7864305,403346383},{456129690,355103696},
{1687155567,848442321},{2125459580,1592370130},{894957909,376067027},{700447433,1116672980},
{565181386,159749077},{1944580492,1160745942},{1196947221,1153872855},{2097148019,2034787288},
{2110779474,2018132953},{1389360550,1719870426},{1785721522,130266075},{144703212,703480796},
{893909335,1810359261},{166198980,533857246},{1611658239,2080048095},{1142421374,1313526752},
{1934619052,289006561},{1166014290,71590882},{2033709281,929280995},{193461903,452531172},
{1949823353,641901541},{219676253,842142694},{46661543,1919058919},{2087186617,1551332328},
{935852295,1563431913},{359136595,1205191658},{1368389073,1482540011},{845150646,1955435500},
{1901064672,1242506221},{1609036804,1518437358},{1048049713,1678963695},{923793694,852562928},
{1220015851,935351281},{570948543,341824498},{1529869471,795411443},{1065351184,1540305908},
{1938813339,764322805},{1084225517,534520822},{771226177,1370190839},{1427109215,333140984},
{1267725966,1259238393},{2119692397,371291130},{1184888620,856609787},{1983902021,1748534268},
{1523053740,213165053},{613415790,1222681598},{1150285678,253785087},{1607988241,51721216},
{89653077,2141955073},{124780306,757888002},{1396700569,1410057219},{1319630382,152396804},
{112197418,1481851909},{1753740015,1363227654},{1194325787,1171219463},{1398273436,1482560520},
{601357189,599942153},{1533015188,2014512138},{1121449893,1449665547},{1461712158,535024652},
{289406424,1154040845},{1747972869,2094470158},{1260910238,1989198863},{1186461482,800470032},
{1241511620,243692561},{1610609666,226157586},{377486640,2138063891},{28311498,1429591060},
{845674931,488002581},{1469576463,175158294},{998242448,1397642263},{149421795,1135215640},
{430439627,544293913},{413138156,761111578},{1332213268,2000004123},{1569715281,1118905372},
{965736654,190239773},{711981747,194626590},{667417354,394400799},{1992290620,129188896},
{1971843411,1045877793},{1259337379,1609147426},{500169798,1785967651},{1115682738,1375720484},
{1134557068,412107813},{1502082255,211678246},{1844441671,426554407},{1748497148,750249000},
{1119352745,700036137},{489159772,2038125610},{1779954380,1375904811},{1430779230,1871877164},
{724564634,1586426925},{252706334,607069230},{87031642,1492952111},{1690825578,854959152},
{272104953,481088561},{1497887959,205796402},{381156649,1520092211},{1459615012,1495258164},
{2127556691,1124439093},{1908404689,1767306294},{1128265624,1239643191},{716700333,1196811320},
{1222112998,462750777},{1881666053,764490810},{1328018972,176669755},{724040347,1639134268},
{478674031,923292733},{1106769858,671638590},{537394175,258360383},{627571541,1297593408},
{506461242,920671297},{286260702,1000670274},{2005397776,942076995},{868219275,1952289860},
{1603269649,651162693},{326630801,1743598662},{808450554,860750919},{1493693663,1532527688},
{1692922724,1758262345},{1773662929,1094141002},{565705674,563307595},{219151966,509551692},
{1917317561,372966477},{411565295,1523639374},{1625813988,207782991},{1270871689,1065370704},
{810023415,1961489489},{343407985,363615314},{2129654047,1230562387},{308805043,581891156},
{1854927408,668820565},{680000239,345261142},{1564996698,2089210967},{1320154668,1202230360},
{2053632184,1094808665},{1506800838,1579689050},{939522305,432993371},{1807741576,1457882204},
{483916902,1012884573},{332397958,586568798},{1953493376,1515775071},{1375729092,2057409632},
{140508916,1403708513},{1314387509,700953698},{1186985769,1746011235},{1346369016,1670485092},
{1611133955,830583909},{2116022489,533460070},{1030748243,1260467303},{1527772330,1529087080},
{276299249,1169224809},{2052583612,135672938},{1174927167,1305179243},{1800401563,517727340},
{1146091382,1779516525},{571997117,1729434734},{1300231760,1331807343},{1253045930,1660384368},
{1336931850,439645297},{1640494023,1185404018},{441973941,2003211379},{324009366,1863468148},
{1962930534,2037089397},{1132984207,1220306038},{811071990,728675447},{169344701,1046434936},
{380632364,1651709049},{186646172,1861153914},{898103631,1973728379},{1029175383,571049084},
{657455898,2145371261},{425196757,1782910},{1452274993,1309717631},{1269823115,2025698432},
{264240648,1656755329},{1538782351,1041196162},{942143740,1475376259},{876083577,697357444},
{504888381,789972101},{750778984,1068156038},{497548364,472769671},{156761813,870380680},
{58195857,1575728265},{1518335154,755164298},{1249375921,634803339},{636484419,681772172},
{406846712,1480369293},{979892403,614352014},{865597838,859329679},{1437070670,1946084496},
{757594716,112645265},{982513841,45515922},{106954548,205649043},{228589132,752784532},
{1159198557,1336657045},{2013262084,147235990},{300940738,1258103959},{1822421623,1142764696},
{1480586488,155870361},{1811411591,1800336538},{1484780788,368403611},{1873277459,1250231452},
{1345320442,161350813},{202899069,1307661470},{1517286578,874800287},{1172830022,42325152},
{1837625936,591643809},{1117779884,1602725026},{1714942797,182469795},{2122313869,646558884},
{649591593,1244488869},{100663104,1573119142},{1720709938,999396519},{135790333,941221032},
{1818751603,370173097},{434109636,962880682},{719321764,1810736299},{752876136,1929704620},
{1923084729,575935661},{491256919,1411560622},{793770518,213718191},{1223685858,1705149616},
{505936955,635077809},{1746924285,1880478898},{1965027692,875455667},{1362621917,829273268},
{1462236443,1283699893},{1232074450,1754887350},{526384148,1612539063},{1120925606,1690629304},
{1285027437,1319937209},{1930424747,521417914},{1232598738,675923131},{968358089,922772668},
{1039136834,11441341},{1053292587,1588786366},{594541458,2026452159},{1514665147,1326159040},
{903346501,1087890625},{952105192,257721538},{1213200121,914371779},{495975502,418759876},
{1827664490,552649925},{663747343,1844741318},{1444410688,727794887},{1695544163,854951112},
{2141189169,2129450185},{1522005163,1522263242},{1495790812,261412043},{1647309755,1245480140},
{822082016,2140959949},{2059923662,364598478},{816839147,1583252687},{1093138398,500982992},
{1155528552,602211537},{1583346740,2135618770},{716176044,1527526611},{1987572032,391693524},
{1475867909,1586910421},{534772740,977892566},{741866106,705459415},{480771180,1855087832},
{319815070,1177081049},{341310837,1416127706},{450886820,1989682395},{412613869,683697372},
{978319542,856057053},{1147664243,284681438},{162528971,1394144479},{807926268,1204782304},
{110624557,2105595105},{2091905158,689808610},{976746681,1680749795},{1043331131,495355108},
{1768944343,976172261},{1604318220,1005077734},{199229060,2033206503},{1498412250,1076684008},
{1723331372,438711529},{512228399,1857701098},{418381026,75453675},{1490023656,555562220},
{1488450796,2075022573},{347602281,922121454},{456653977,809407727},{1068496908,531666160},
{1942483340,1092773105},{849344942,882615538},{1891103221,867140851},{2066215129,1714427124},
{651688741,1793758453},{1386739117,920528118},{87555929,586450167},{1561850973,1931343096},
{463469708,1332974841},{1897918941,4654330},{1238365894,908047611},{1460139298,986084604},
{692583127,1543005437},{516422696,752772350},{663223056,1298928895},{1425536361,1180984576},
{132644611,1310655745},{579861423,2106033410},{1973940556,1487406339},{1797780132,284026116},
{53477274,1063113989},{51380126,1916880134},{1671951245,1887479047},{1427633506,1319175432},
{174063284,867276041},{1124071328,1159382282},{794819092,81253643},{1789915818,1310418188},
{99090243,1342760205},{1264580247,1096992014},{117440288,782157071},{1185937194,1393661200},
{692058841,423556369},{557841368,1198802194},{479198318,1773282579},{912259380,1478788372},
{1635775442,1184527637},{12058601,1563874582},{413662443,1462871319},{1445459259,1770251544},
{141033203,1232147737},{246939177,664302874},{1020786790,686400795},{299367877,1006568732},
{1513092283,1475511581},{1872753167,938824990},{95420234,206050591},{1333786128,1467544864},
{542112760,651404577},{1483207926,890254626},{758119002,670827811},{1698689880,1940841764},
{2088235223,159552805},{1706554193,1746322726},{1937764756,785315111},{803731971,485004584},
{654834463,268457257},{1941434764,80069930},{49807265,62620971},{306183608,1990571308},
{1573385288,964592941},{1326446114,502359342},{877656438,951833903},{1710224195,31147312},
{424672470,1686697265},{352845151,941286706},{1537733771,1113859379},{2095575178,199910708},
{1691874149,1479882037},{962066646,719017270},{1631581145,2111288631},{234356289,1865274680},
{37224377,102597945},{1131411346,484730170},{1252521643,1411573051},{854063523,1858213180},
{1339028998,127255869},{743438967,1902200126},{326106516,1429955903},{1039661122,515945792},
{194510477,2066986305},{759167576,328897858},{1369437644,2133468483},{84934494,1778320708},
{216530531,1509221701},{88604503,1418192198},{1862791712,705123655},{924842268,1304380744},
{925366555,1699833161},{1078458359,1467995466},{671611647,2032780619},{643300150,1175926092},
{337640828,235095373},{693107414,1764971854},{5242870,1261516111},{894433622,111748432},
{1593832480,2058732881},{487586910,719512914},{1859121712,551794003},{1547695225,1731319124},
{871364995,1034790229},{506985529,145540438},{1195374362,376382807},{1164965714,1356690776},
{1929376170,625821017},{3670009,1274557786},{1264055957,310818139},{530054157,1199760732},
{390069528,476591453},{1312814649,848172382},{2008543497,853554527},{2138042848,489506144},
{1680864137,426169697},{863500689,1529283938},{1692398440,585426275},{1545073790,52426084},
{1356854758,1308157285},{72875893,595645798},{192413329,1916626279},{161480396,207279464},
{604502911,771822953},{38797238,2010576234},{2034757878,1893549419},{574618552,1159284076},
{913307954,748176749},{1871180318,226616686},{449313960,342594927},{1189607204,1266447728},
{258997779,1380861297},{1695019883,1482241394},{175636145,260060531},{630717261,1218262388},
{1945104773,1893606773},{2136993957,721012086},{2118643802,2109638007},{1903161821,2004747640},
{23592915,1130513785},{423099610,2011719034},{966260941,105895291},{1989144918,733861244},
{1737487119,1632249213},{2093478013,429364606},{2144861228,2108163455},{357563735,551445888},
{774371899,226444673},{1711272774,2086286722},{152043232,2092299651},{1704981326,545453444},
{2102390898,1187218821},{70778745,1259541894},{33030081,1951466887},{1514140857,486360456},
{1744302853,556414345},{1060108315,333473162},{466091143,1002804619},{810547702,2065429900},
{1466955027,2108687757},{700971719,1129059726},{1449653557,285468047},{919599398,1691497872},
{1779430080,1067578769},{1590162473,1141020050},{1842868826,57308563},{1245181628,2027861396},
{1513616571,500716949},{1688204143,484517270},{1044379704,842962327},{1227355868,1861813656},
{550501350,369108377},{988280996,1106179482},{2054680754,2136323483},{204471930,143717788},
{2020602106,1206863261},{571472830,944502174},{427818192,378189215},{1669329809,455988640},
{961018071,1418110369},{1455420712,528971170},{824703452,1048561059},{1684009860,437134756},
{376438066,891942309},{14155749,1689150886},{335019393,2092406183},{878705012,2138989992},
{1385690542,1936291241},{319290783,607114666},{1178597177,1542309291},{1361573344,1970877868},
{2024796419,526308781},{572521404,82908590},{444071089,489559471},{1837101650,945337776},
{345505133,622069169},{116391714,1103177138},{1239414468,5096883},{1234695887,1518921140},
{1440216391,2040460725},{224919123,706045366},{2139091024,244553143},{1990193479,1686320568},
{1866461723,763528633},{1225783007,1725568442},{1161295707,1496663483},{181403302,1144329660},
{1075836924,1097045437},{1950871956,1718388158},{995621013,1287574975},{1998582061,1660925376},
{364379465,612615617},{996669588,1689081282},{2086662267,183932355},{56098709,817837508},
{639630140,121427397},{1125119902,1191413190},{17825758,1458341319},{1832383074,996767176},
{145227499,1310021065},{709360312,1224852938},{1618998260,263538123},{1189082917,296203724},
{1010301049,691467725},{1485305078,862578126},{792721946,1061983695},{931657999,1986876880},
{490208345,1248175569},{927987991,938120658},{1444934977,97805779},{1889530353,814347732},
{2042622158,1677211093},{1059584027,397186518},{1205335813,625075671},{2067263692,1201817048},
{1148712817,820442585},{867694986,1788036570},{1309668926,1045714395},{1038612553,929592796},
{1641542615,1122977245},{56622996,11711966},{1787294389,2031945183},{1092614108,1361679840},
{1434449235,1296045537},{1949299076,1581209058},{467664005,1408267747},{1749545725,657405412},
{593492884,927704549},{1886908924,520152550},{1654125485,1133528551},{1959260521,700028392},
{450362533,1966597609},{201850495,1918273002},{2030039293,2064152043},{479722605,1124877804},
{2074079402,1757898221},{294125007,1638413806},{452983968,282195439},{2131226750,956982768},
{1410332034,1568351729},{1155004261,883308018},{1422914924,1819301363},{397409546,2073662964},
{1137178504,548476405},{1384117686,1791206902},{1072166915,2036581879},{1327494685,1098937848},
{1650455487,558581241},{1855975985,1584182778},{1746399998,2091398651},{356515161,614426108},
{1605891082,1736340989},{1121974184,1675183614},{2140664054,1489364479},{600832902,1327101440},
{1334834702,1761744385},{1303901770,352192002},{439352507,773146115},{740817532,1333786116},
{2078273684,966063621},{113245992,1777116678},{1708127050,29550087},{1288173159,910611976},
{147324647,555410953},{954202340,269100554},{1471673612,1120966155},{1802498709,1362556428},
{1177548608,347928077},{1088419813,1181455886},{914880815,2057094671},{760216150,1970869776},
{2054156489,1200305681},{1892151784,868029970},{1741681417,74810899},{1165490003,333411860},
{1296037465,923837973},{776993334,1591567894},{2043146452,1356523031},{780663344,420632088},
{1628435422,1073198617},{401603842,95413786},{1796207265,201307675},{1600123925,1081955868},
{1771041487,1666266653},{421526748,775370270},{2119168056,1256687135},{433585349,1522632224},
{1314911796,1851078177},{671087360,1530709538},{1047001139,1410700835},{1940910479,1353025060},
{444595376,1607644709},{1500509396,841524774},{983038126,1313392167},{2082992257,339637800},
{2027942152,1047422505},{222297689,1717057066},{1275065985,1549190699},{1116207027,90101292},
{599784328,1116087853},{1429206364,1563080238},{1631056857,134600239},{1875898893,182670896},
{1743778564,33117745},{203423356,1200260658},{1449129268,481060403},{159907536,859092532},
{1007155328,2039064117},{2068312235,948487734},{955250914,2043070007},{835189191,1779476024},
{82313059,1982346809},{747108975,466245178},{1626338278,1422714427},{2051010760,492926524},
{2105536656,391280189},{1115158450,1982772798},{1966076265,1899210303},{2048913605,1311352384},
{501218372,969008705},{1663562653,871269954},{618658662,1261598275},{445643951,937408068},
{1210054396,529843781},{1636299734,1503876678},{2006970637,1261901383},{1922036142,797324872},
{487062623,284034633},{1132459920,989840970},{438303932,1287804491},{1473770759,2059626060},
{500694085,1998734925},{190840468,1281082958},{1507849418,559793743},{605027198,1912489552},
{1076361211,563521105},{251133473,1843144274},{1679815549,962393683},{1013971060,1304442452},
{1114634162,633357909},{1907880393,262800982},{1702884177,1232238167},{829946323,928716376},
{11010027,503961177},{361758030,2096178778},{31457220,96249435},{477101170,211621468},
{1619522543,1378809437},{947386610,1292813918},{1367340498,1729283679},{1625289710,1359890016},
{541064185,377951841},{681573100,2058864226},{1051719722,722953827},{1045428282,1221080676},
{556268508,1800943205},{1584919602,90130022},{465566856,1096844903},{2046292177,1148372584},
{1052244011,1007289961},{1663038367,1977947754},{1850733110,712463979},{277347824,1863001708},
{1989669199,82151021},{1721234222,3192430},{742390392,1034810991},{2112876615,2129675888},
{1639445466,204211825},{976222394,1289094770},{2132275404,1028638323},{1448604984,1608599156},
{17301471,268519029},{313523626,479876726},{350223716,431392375},{535821315,318621304},
{1011349624,251135609},{597687180,1484879482},{210763374,1183438459},{122158871,1296623228},
{310902191,158439037},{618134373,1074099838},{1221588711,1947481727},{1579152445,1364940416},
{1198520082,1450514049},{2037903588,211334786},{71827319,9905795},{1896870366,898713220},
{1031796816,1905747589},{1780478652,1683523206},{1788867248,1775048327},{2003824914,1447462536},
{953678053,1254885001},{875559290,1273128586},{921172259,1973417611},{707263163,1035241100},
{80740198,475870861},{20971480,1977665166},{767031882,1175414415},{1803547289,818538128},
{1377826236,1021249169},{1094186969,2147395218},{688913121,1349076627},{1341126148,1765557908},
{379059501,1621919381},{1407186319,451991190},{354942299,1669285527},{153616091,1894061720},
{1896346080,1758660249},{69730171,572155546},{587201440,1327113883},{138411768,462022300},
{1434973519,1175013021},{1658319782,178558622},{168820414,1128388255},{658504472,1039574688},
{255327769,767080097},{255852056,567953058},{1772614368,959555235},{164101831,394380964},
{351796577,433190565},{152567517,119035558},{2080370829,1488918183},{1250424496,1850545832},
{1190655778,1340761769},{439876793,2080200362},{2132799614,1883612843},{1522529448,371218092},
{126877454,966952621},{1478489341,1752647342},{946338035,27981487},{862976404,480122544},
{308280757,1669105329},{57671570,1853478578},{1894773221,1602778803},{1809838729,332940980},
{455081118,501118645},{696777423,1488271030},{1559753825,348473015},{1116731310,717444792},
{1736438553,1887401657},{1594881068,492418746},{1854403132,482100923},{401079555,1237706428},
{1329591835,1388787389},{1301804622,1342617278},{2064642207,593938111},{773847612,1642743488},
{2092953730,570980033},{114294566,986859202},{227016271,1098012355},{2085089429,275928772},
{1704457046,1904424645},{1475343618,1504757446},{1553986668,1493186247},{900725066,2112943816},
{1442837824,2085996233},{1193277213,257869514},{63963014,961275595},{1017116780,2058413772},
{189267608,1274717901},{1040185412,899507918},{349175142,582043343},{1546646650,2077185744},
{303562173,223098577},{1504703694,1169921746},{1103624136,1831552723},{603454337,2065979092},
{800586249,193324757},{984086699,94709462},{318242209,464152279},{759691863,805918424},
{1482159349,152717017},{748157549,727906010},{375389492,1937360603},{2030563558,2143753948},
{1927279014,913463005},{646445871,255993566},{1560802399,947193567},{1980756291,1503246048},
{482344040,757061345},{467139717,1339668194},{147848934,1237599971},{1642591183,4646628},
{778041910,251373285},{2016407862,2010089190},{1031272530,1628620519},{934803721,1354200808},
{231734854,1609430761},{1850208829,1912358634},{2057826476,840967915},{699923145,517289708},
{228064845,2096207597},{379583788,1412593390},{333970820,1703458543},{764410449,200374000},
{334495106,1813153521},{240123446,156317426},{1633154005,568620787},{1877996041,147478260},
{771750465,541132533},{1923609010,271218422},{1531966614,2046793463},{1515713719,518280952},
{2101866584,1182234361},{504364095,356869882},{848820653,534546171},{1143469947,86001404},
{419953887,777000701},{621804382,2105743102},{1401419156,102184703},{886569317,2081810176},
{745536114,328922881},{812120563,1082877698},{200277634,1415190275},{124256019,1226856196},
{1216345843,960472837},{1697117020,1347843846},{288882137,2105648903},{186121885,214857480},
{2135945565,1613817609},{1592783910,1475495690},{1825043052,1327204107},{1470100749,973215500},
{1775760075,1024005901},{1001388170,1151596302},{2079322339,1543960335},{1921511859,136926992},
{623377243,552695569},{1021835363,1582884626},{1182791472,1579624211},{1218442989,1503516436},
{2110255290,1367549717},{1852305975,746243862},{302513599,174003991},{778566196,1809905432},
{940570878,216024857},{2001203506,731117338},{1081604081,841500443},{1061156888,69768988},
{1627911137,1229498141},{1868558873,881559326},{44564395,925648671},{2076176526,230663968},
{741341818,664971041},{314572200,1137288994},{1910501831,154990371},{1911550403,220501796},
{1998057778,442648357},{60293005,2012542758},{560987090,2096404263},{27787211,172750632},
{2088759465,1102231337},{1082652655,492193578},{1472197901,1697199915},{2141712925,1222936364},
{666893064,237995821},{746060401,407971630},{62390153,74061615},{1230501589,570869552},
{1245705919,351684401},{844626357,840738610},{676854517,222590771},{230686280,48809780},
{631765836,486958901},{273677814,1471432502},{475004024,1166411575},{802159110,1339246392},
{469761152,646842169},{1856500276,1715296058},{257424917,1329628987},{1313338944,2067687228},
{131071750,581756733},{1202190093,1217779518},{1499460820,811472703},{661125910,229246784},
{1474295047,173696833},{1730671392,1528346434},{1796731556,913299267},{1845490243,73193284},
{735574661,1368803141},{1350563313,835770182},{1613231099,1333000007},{1795682975,1652311880},
{383778084,496404297},{360709456,2111354698},{2055729362,1756452683},{1982329196,285939532},
{1599075350,1200117581},{95944521,1213437774},{1049098287,351487823},{1054341158,567056208},
{1550316659,1188636497},{949483758,396363602},{149946082,1313945427},{886045031,2130360148},
{1347417590,265254741},{1572861,1851017046},{1776808644,1506801495},{1150809967,1350338392},
{307756469,295606105},{2059399345,139888474},{736098948,397047643},{717224616,1597871964},
{1491072229,652728157},{1170208584,989546334},{151518943,1784543071},{2128080950,301229920},
{62914440,1342437217},{471858300,951920482},{516946982,1157576547},{488111197,1199535972},
{25165776,1745925989},{2090332306,680417126},{967309516,1557559143},{1972891981,1613625192},
{1096808404,1839535977},{342359411,2016274282},{484965475,1758619499},{677903091,862902124},
{892860761,646555501},{1433924950,731297646},{1463285021,1868879727},{522189852,1689560944},
{311426479,1070401393},{2128605418,2067515250},{316669348,1938782067},{1883763204,378550132},
{568327108,1268938613},{581434283,1762932598},{916453676,70526839},{705166015,1634965368},
{1459090721,513828729},{1071642628,199325562},{396360972,1045546875},{435682497,1692731260},
{1015543919,1280071549},{2063069353,1729628030},{1789391535,51353471},{185597600,225286016},
{1539306641,836679553},{1914171847,757503874},{1699214174,704386947},{358612308,783390596},
{1879044608,743405445},{1073739777,448542598},{1568142424,666253191},{2045767879,739116936},
{1197471511,748627849},{556792795,1656895370},{710933172,294762379},{755497567,2063824780},
{1251997356,850184077},{1792012968,31315854},{1325921826,1317271439},{1391457699,1070884752},
{1048574,497276817},{544209906,270907282},{624425817,943167379},{374340919,219346836},
{343932272,2096674709},{1338504712,1834842006},{662174482,829122455},{370146622,187455384},
{198704773,242722713},{1717564215,1523709850},{2106060952,1843025819},{179306154,2143209372},
{1200617230,791709597},{710408885,1614727070},{2011689246,1957590943},{817887721,1840924576},
{920647976,835348385},{128450315,677697442},{1127741337,1323829155},{1361049060,569866148},
{400555268,1606682533},{130547463,835856294},{821033442,1742370727},{1032845390,1876391848},
{1718088499,975194025},{1830810206,1234798506},{1420817772,1190442923},{1981280605,600070060},
{465042570,1354413997},{1228404441,322467758},{168296127,775112623},{620231521,1954031536},
{717748903,1002866609},{515898408,271091634},{1808790164,1335977907},{1638921167,1385785268},
{1505752264,1365292981},{1918366138,1595180982},{1052768298,1687693239},{1113061301,864044984},
{2134896770,1574791097},{722467488,330028986},{1725952804,952424379},{1512043710,1126381500},
{827324886,2057975741},{1510470849,1469921214},{833616330,1224509375},{2044719312,815327168},
{1418720630,501827521},{1756361456,1533581250},{1558180966,1650980803},{616037226,2041382852},
{784857639,1968277445},{750254697,1477949382},{1104148422,7395271},{513801260,1676216264},
{57147283,764069833},{29884359,99166154},{791673371,1746024395},{1919938994,1869502412},
{1168111436,1104635853},{1145042809,2021808078},{1866986010,903096271},{2023223607,323282896},
{776469047,1915471825},{517995556,1997907922},{934279436,1043269587},{386923806,146978772},
{1383593393,2039990229},{1071118342,1080608726},{1206908678,2093332439},{1864888859,1218140120},
{649067306,804394969},{1086322664,2099914714},{1954541956,967649243},{388496667,1669507036},
{801634823,419346397},{1735389981,1073055710},{1199568656,1398818783},{634387272,620513248},
{2106585222,599439329},{617610087,175687650},{2109730918,1430255587},{798489101,25429988},
{2044195027,1081821157},{1281357429,1016346598},{1967124825,1003898855},{1678242687,1096226792},
{801110536,1921554409},{448789672,1951143914},{1538258059,1898366955},{1932521896,1436592108},
{554171360,1214576621},{1899491814,2130651118},{1550840949,1091098607},{366476613,260536304},
{363855178,1960712177},{848296366,1222514674},{592444311,1693415411},{869792133,749766644},
{1587016753,902019061},{1918890438,98080758},{1857548845,1282541559},{1579676732,919046136},
{433061062,303359993},{1049622576,560191482},{292027859,1434994683},{353893725,59553788},
{1111488440,2004953085},{2142761248,849893374},{1175451456,340994047},{1212675834,1234057216},
{1033893964,543610881},{601881476,1335851010},{1097332692,1061971971},{1831858778,1676494852},
{1261958809,911759365},{542637046,567629830},{1573909575,36952071},{1794110117,301471752},
{1835528795,294524937},{805829119,1671837706},{1843393102,555456523},{327155088,164517900},
{424148183,1749760013},{218103392,1376311310},{1530393755,1045026831},{720894625,161744912},
{165674692,499333137},{1575482437,458524690},{1762652909,1950271507},{2009592073,1620211732},
{4718583,944392213},{220724827,1109723158},{1291843168,254650391},{369098048,4610072},
{873462142,219899929},{125828880,223823898},{1847587390,638867483},{1408234884,1743259676},
{1323300389,19458077},{764934735,1320093726},{936900870,1878501407},{1443886402,1162688544},
{1310193213,660850721},{678427380,125159458},{309853617,835647523},{2017456377,1117460516},
{1735914262,1505146917},{246414890,1394530342},{1421342058,1315125287},{1460663586,1274947624},
{1364719063,1955698729},{1786245809,432871466},{1193801500,1386145835},{22020054,684165164},
{1463809304,246708269},{1675621253,1448785966},{1764225759,582764591},{667941639,998746160},
{1754264314,520173617},{1673524105,1297922098},{576715700,127576115},{1158674272,787937332},
{650640168,295626805},{2123886660,1410533430},{1730147101,936581175},{1612182525,480004152},
{242220594,1620072505},{469236865,2136467514},{1534588049,592021563},{1240987329,1962154044},
{1633678294,377305149},{336067968,423991358},{968882379,1170524223},{1694495587,35153984},
{2034233571,1828026433},{1462760731,1884604482},{1102575561,908539971},{157810387,1986271300},
{723516060,1697110085},{1782575804,689268806},{728234643,1233537095},{94895947,1763473480},
{1360524774,1755306057},{767556170,492726346},{120061723,1814808651},{1852830258,651651148},
{1173878593,305080397},{1716515638,1841240142},{971503812,464103503},{1337980424,1552238672},
{2086137981,32426065},{791149083,577312850},{395836685,1836148819},{597162893,127551572},
{709884599,135481429},{1015019632,1309001814},{1302853196,1791559767},{503839808,1615579224},
{41418673,1321887833},{2116546717,942090330},{1635251159,1985472603},{38272951,189847644},
{1772090061,1239976029},{1705505611,1378801758},{158334674,716920927},{1255143079,935581792},
{173014710,1262835809},{2105012344,1616271458},{358088021,122120291},{1601696785,2105428068},
{875035007,2081130597},{1865413150,1355348070},{775944762,187168871},{2022174975,1522735208},
{1945629070,931391593},{136838907,833177706},{442498228,290117739},{781187631,1237346412},
{807401981,373434477},{1691349863,493250670},{1311241787,1337596015},{499121224,1684633712},
{1815081597,432003185},{1825567341,1637709938},{1225258719,952105075},{2019553534,384374900},
{996145302,1096194165},{1693971299,839952502},{1062729751,1935726711},{1501033684,915728504},
{772799039,404994169},{230161993,1295501434},{1235744463,2047010939},{804256262,644294780},
{103808826,1109051517},{1139275655,1038784638},{2025844985,1106110591},{1859645994,1948223616},
{486014049,611207297},{257949204,1763260546},{991951005,1496127619},{1441789254,730683524},
{1901588950,1566156933},{1781002940,689563782},{321912218,786835591},{1397749146,1749690504},
{702020293,1664211081},{2007494939,1805564042},{1396176281,1870604427},{13107175,429648012},
{816314859,619911309},{719846051,1759987854},{1765798617,91986063},{579337138,584099984},
{160431822,708872337},{698874571,1274456210},{1162868566,1488394387},{1243084477,735819924},
{1263531670,2022877333},{1044903994,301426838},{2143810008,650172567},{1937240471,408246424},
{1769992931,858142873},{1997533476,617855130},{1956114818,1309915291},{817363434,313563292},
{2103439463,156412061},{5767157,402938014},{233832002,1411524767},{637008708,1276774560},
{1916268986,1957832865},{1289221733,2005498018},{969930950,455772323},{528481296,1397991588},
{1008203902,1651599525},{896006485,897632422},{263192074,1276807335},{104857400,1332885672},
{1904734685,459274409},{740293244,1261590698},{839907774,1559378091},{1283454577,1534847148},
{788527648,591317165},{933230861,645114030},{182976163,1196882095},{1075312637,1581082800},
{812644850,14895281},{378010927,758143154},{2010116364,1333237939},{1476392194,2006431924},
{393739537,295905461},{321387931,913111222},{126353167,347306167},{1939861911,1439566008},
{1615852534,1495271609},{1140848513,1200396474},{27262924,1898231995},{1258288800,420161724},
{1519383727,429729981},{2089283721,1505908926},{625474392,1396435135},{408419573,1448384704},
{2113925195,1093699777},{920123685,1632123074},{586152867,600197315},{1346893303,271268036},
{1125644189,1383889093},{216006244,233539782},{499645511,1087645895},{1179121464,366418120},
{795867668,916224201},{1326970397,430655690},{1617949682,1199970507},{1976561995,1228839116},
{251657760,996055245},{972028098,901986510},{1495266526,479725775},{790624796,471124176},
{67108736,817895633},{1192752925,639338706},{43515821,520063187},{1069021195,1626966228},
{1370486219,364357845},{621280095,829135062},{103284539,267282647},{532151306,975657176},
{25690063,204142809},{2026369291,2058995930},{689437408,681720027},{382729511,1694861532},
{800061962,816404701},{1110964153,2129000670},{291503572,1534695647},{173538997,2018322656},
{1567093844,1704339681},{1621619696,969001186},{981465264,1414957283},{473955448,877312228},
{962590932,215656677},{204996217,903026918},{2018504968,359667943},{738196096,1265674472},
{508558391,173619433},{2043670764,548858090},{2082468156,632785131},{640678716,965339372},
{448265385,1400961261},{1509422277,573372654},{652213028,377829615},{262143500,1211885808},
{1451226417,1509288177},{1253570220,1564408050},{585104292,725911795},{1331164693,1548282100},
{595065745,322545909},{1925706151,1214302454},{756546141,754813175},{1904210394,2098215160},
{45612969,1213397241},{631241548,259524858},{1660416933,272173307},{602405764,112654588},
{732953227,938010877},{489684059,1646778622},{766507595,1845631231},{1034418251,672700672},
{972552387,267397377},{1920987571,107198722},{1756885737,581175555},{402128129,1151916292},
{272629240,867436805},{606075772,224954630},{1397224864,198666503},{1774711495,242993416},
{1891627496,1676110089},{1404564873,2048514314},{446168238,164632843},{367525188,555698444},
{1557132394,1225623821},{1925181892,1289316622},{385350945,193693967},{1349514738,440060176},
{1445983546,204822801},{443546802,373852434},{1059059741,475269395},{338689402,190028052},
{363330891,132077845},{66584449,873564438},{1602745369,1141786903},{317193635,1735571736},
{633338696,1686800665},{1164441427,470206746},{31981507,466057499},{582482857,1165924636},
{1004533892,550459677},{1985999188,792082718},{1053816870,303016223},{1140324225,1302767904},
{1042806843,1487604001},{1784672950,1274292514},{498596938,1416055075},{534248453,1103952164},
{1405613449,2112629029},{1994912046,720415014},{262667787,717371687},{1227880154,1656113448},
{390593815,454465833},{551549926,769333546},{628620113,12937515},{815790573,2018896172},
{751827559,1274018093},{1934094773,1521273134},{1353184751,627878191},{2058350786,634882352},
{1770517203,812149041},{137363194,1855494450},{196083338,204044595},{77594476,883894580},
{1355806182,1925269813},{404225277,942565686},{1064302610,630307127},{1719661367,1060956472},
{2089807984,1153112377},{1123547041,1791228218},{522714139,51145019},{2014310656,1577687356},
{1358951904,319834429},{156237526,528800062},{948435185,1323878719},{1231550163,1116813632},
{1556608104,1724549441},{1036515399,1386674498},{2121789529,496994627},{240647733,1218763076},
{1978659149,1473075525},{65535875,1921042758},{1665135528,1052678471},{1712845645,499038536},
{167771840,714303817},{164626119,673077578},{1254618796,464132427},{54525848,1944598860},
{1880093183,20166989},{1681912697,1819941198},{76545902,1499167055},{994572441,1609202000},
{1722282801,1591212369},{1420293483,953542994},{1088944101,2111482195},{1149761393,114182484},
{373816631,94038357},{1298134612,1483884886},{109575983,1741551959},{821557729,1389062488},
{650115880,1231812953},{819984868,2131880282},{1248327348,1806117211},{184024737,772790620},
{1883238911,1078937949},{1638396878,1852713310},{1357379046,385390943},{64487301,431823200},
{1876423174,1272428897},{664795916,1107462498},{134741759,1934981475},{1750594296,1892768100},
{1149237104,2134419813},{1610085384,713177446},{897055057,8083815},{2004349213,1813789032},
{127401741,274282857},{586677153,168348010},{842004922,2021497195},{619707234,1111071084},
{297270729,507439469},{1680339836,209967470},{1006631040,750754159},{666368777,1241643376},
{1332737554,1354062193},{2107109470,1889540466},{234880576,1261369715},{852490663,1433352564},
{918026537,2032327029},{1290270307,1845594486},{1060632603,2111900023},{2047340744,247671160},
{1233123025,47196537},{855112097,647833978},{1545598086,547629435},{1481635063,1300904316},
{1948774794,1071806845},{1821373040,901929342},{1541928069,1081952639},{344456559,10971520},
{936376582,2035227009},{35651516,1540876674},{317717922,563063171},{384826658,490072452},
{143654638,1951783301},{1798304414,219060614},{559414229,1672882567},{1122498467,759486856},
{543161332,611547529},{2041049311,820517258},{1436546380,2006276491},{2069885095,438864268},
{495451215,610769293},{2136469684,2073692558},{851442088,1853376911},{1025505373,1164319120},
{357039447,194722193},{1102051275,54102418},{476576883,1172273555},{1909453259,471419284},
{191364755,1516849557},{1760555747,1359366550},{167247553,1901914519},{1304426064,1349507480},
{1421866346,405531033},{1019213931,523876762},{1382544820,989886875},{1236793033,2013632924},
{254803482,435714461},{1745875716,1614252446},{662698768,1338427807},{1913647557,1763633568},
{2083516542,529136033},{416808165,134343074},{1379399097,2009737635},{1893200384,741685668},
{250609186,579316133},{1508373700,1494432166},{1720185647,1273592231},{1183315761,721160616},
{754448993,106060201},{1064826898,1352194474},{1324348962,1913809323},{298319303,2014624172},
{2002776342,130849197},{15728610,906451374},{600308615,133564847},{1464857882,638593456},
{813169137,28326321},{336592254,410614194},{2121265205,1068779955},{2037379333,1761159604},
{1532490903,893270453},{96993095,1758771638},{828373461,1222003127},{784333352,1891580344},
{157286101,2132486585},{594017171,483604922},{28835785,175372731},{376962353,1316239804},
{472382589,1167251901},{411041009,512821694},{1975513471,1479854527},{1745351429,1734015424},
{247463464,2104605121},{648543020,267041218},{2042097942,1108797891},{1119877032,1750563268},
{1566569558,2030623173},{415235304,1747642822},{2005922081,1008990663},{1261434522,77781448},
{864024976,262621641},{901249354,1164609994},{1129838485,1801480651},{454556829,1420388812},
{630192974,2135058893},{1358427618,898795982},{1725428517,634067407},{1467479321,150084048},
{440925368,1279588817},{1801450138,1962441170},{674757369,1165572563},{481819754,1522481620},
{1203238668,1045748181},{1080031221,1868057046},{1330640408,565660119},{245366316,540752344},
{806877693,1906792921},{1816654456,2077346266},{624950104,85780955},{1626862561,1423509980},
{1176500028,2144131549},{753400419,400292318},{470285439,20851167},{72351606,1745828320},
{209190513,466905569},{297795016,1523788258},{655883037,965032419},{883947882,2065316324},
{1984950605,245545445},{1739059979,1539049958},{1271395976,1323084263},{1671426962,2024036840},
{1364194774,806361577},{1623192557,1572911594},{1759507179,1800378859},{51904413,1629370860},
{408943860,2075572717},{1860170310,1334311406},{1605366795,784308719},{657980185,29428208},
{1423963494,388438513},{2067787980,78395890},{945289464,1114700275},{998766735,2020166132},
{865073550,542345717},{1853354559,1382357494},{954726630,643312119},{1416623480,1207253496},
{884996456,220645881},{1384641967,1368594938},{1589113897,1872079355},{675805943,1906924028},
{1414002041,938322429},{1734341396,2107349502},{120586010,1400220159},{831519182,827079168},
{1348990456,1836046849},{872413570,1173559810},{1977610568,1186507267},{1721758509,111897092},
{1275590273,1708435973},{1701311322,1974372870},{1567618132,308132359},{910162232,1022355976},
{707787452,1851218441},{1105196996,1945205258},{1877471750,1515731467},{50331552,1419557388},
{1992814944,521890317},{686291684,2083641870},{548928491,754735631},{2146956941,1526475280},
{318766496,1995082257},{832043470,706451986},{1151858539,1137216019},{1157625701,830159380},
{426245331,448715285},{2108682448,1818802710},{1492645089,102890007},{1505227980,1767438872},
{922745120,1713130009},{1030223956,1790282266},{1028651097,2133752347},{1061681177,1783917084},
{690485979,295160349},{796391953,576735774},{797964814,341875231},{546307054,74152480},
{793246231,1981352481},{472906875,1918798370},{765459021,1355938339},{1450177846,1930803748},
{502791233,1528564261},{570424257,700336678},{1806168723,865204775},{1731719969,1860831784},
{1768420060,29895209},{1156052836,826051114},{1755312878,779299371},{1926754729,1923320364},
{195034764,977320493},{558365655,869341742},{1074788351,1170201135},{739244670,1037736496},
{653785889,2002827825},{1363670487,821254706},{1985474870,1433541171},{808974843,2024294964},
{1812460169,716958261},{462945421,1934408246},{564657100,249543223},{223870549,1800837688},
{1492120803,1902733881},{1963454830,1483799098},{138936055,427575867},{170917563,1769675324},
{904919362,1580243517},{562035664,2000665150},{908589371,788789823},{1000339597,1835145792},
{684718822,2102041153},{1930949033,716958274},{1539830920,958212675},{1584395316,1892825668},
{1941959054,1720613445},{288357850,511101510},{1761080034,1010297415},{293600720,1238645320},
{964688081,1912535625},{653261604,533723722},{1217394414,962689611},{1797255855,22264396},
{1086846951,1942514253},{1248851637,620780110},{90701651,505940559},{985659561,920554064},
{1993863472,1392593489},{391642389,47118930},{2071982229,293886547},{768604743,602765908},
{762837585,1590803029},{184549024,521783894},{938473730,885381719},{861403541,1329760856},
{1535636623,852265561},{1557656682,1698261594},{48234404,739334747},{323485079,222804572},
{738720383,201235037},{40370099,1515879006},{762313299,631642719},{214433383,1340238432},
{1130887060,1387260513},{1378350523,1255524962},{1426060640,2083273315},{1674048398,1321097828},
{1163917140,1719429733},{614464366,1299475046},{1836577366,1600350823},{63438727,813845096},
{1674572679,871783017},{729283217,217979498},{1875374602,1462028907},{672660221,1865042540},
{2122838212,764332653},{1284503150,1512475246},{1645736905,1573546607},{285736415,1811978864},
{2065690823,164665969},{311950765,998247026},{2078797971,507472499},{625998678,1684503156},
{478149744,1044425333},{695204562,1852463734},{223346262,1247038071},{1840771661,1311586936},
{897579344,785775225},{1582822453,555301498},{629668689,1734851195},{647494445,469695100},
{1195898647,254405245},{1954017658,1459161726},{66060162,1094687359},{436731071,1716742784},
{1512567996,2076473985},{641727288,863304322},{398982407,348678787},{385875232,2088155780},
{609745781,1659087493},{533199879,559409798},{814217712,1672624775},{545782768,290605704},
{1935667619,1661672073},{36175803,135731850},{632290122,524946059},{520092704,874523276},
{890763614,71305869},{113770280,258919054},{861927828,172370575},{1817178744,2080443024},
{1267201679,2136857233},{1260385948,502512274},{1376777662,1457769107},{327679375,1659120276},
{1924657586,1321945749},{818412007,1748130454},{945813748,1581742743},{60817292,1189509784},
{554695646,915561113},{891287900,1398114970},{256900631,740625051},{827849175,1521539740},
{1343223294,583240349},{881850734,709696158},{256376344,771058335},{680524527,2127600288},
{1664611225,1280359073},{2053107903,1696410274},{584580005,1343154851},{123731732,1505188516},
{2015883519,33565349},{706738876,1664527014},{1849684545,2057128615},{1583871032,733977256},
{463993995,133024425},{19922906,1343220394},{642775863,1328343723},{781711918,684141228},
{551025637,1717279405},{354418012,794684078},{1993339189,1680067247},{1834480213,571628208},
{295173581,2025360049},{1826091625,1471621810},{1816130180,619989683},{1433400658,1114827444},
{1553462385,849926837},{1316484661,1415346870},{1136654224,1810975415},{337116541,679176888},
{1077934072,239332025},{2114449503,2028370618},{2036330761,1694153403},{779614770,791947964},
{1141897086,1347619517},{1991242044,864381630},{678951665,710978239},{880277874,229329600},
{1776284356,628271809},{853014949,1827896002},{794294805,923552451},{1157101409,367860420},
{417856740,1371486917},{595590032,970488518},{361233743,924056263},{2079846634,819157704},
{1452799277,1227954889},{803207685,134113994},{169868988,492444363},{452459682,857623244},
{691010266,411433677},{1744827143,78170830},{1249900209,256289487},{1555559532,572820176},
{878180725,1568942801},{999815309,1996278482},{1690301288,384129747},{163053257,410630868},
{109051696,394779349},{1793061540,1053600470},{1318057521,492243671},{40894386,147770072},
{1231025876,945081049},{1333261841,10918618},{309329330,1571056347},{609221496,567184092},
{1890054636,1925765853},{1418196335,385309406},{734526087,1104018143},{548404202,594447072},
{1431303514,1201548001},{244317742,1005382370},{1163392853,1414154979},{1406137734,476531428},
{218627679,631945957},{1234171598,2018568934},{2040000731,1057589991},{1091041253,683391720},
{846723506,1284696809},{1440740676,1234676458},{143130351,1616911083},{333446532,341002988},
{1503130831,829610733},{1977086283,1107393262},{837286341,1001347823},{2081943721,1629174512},
{880802160,85330673},{1279260281,1406331634},{1905783246,588901107},{1067448332,761617140},
{104333113,1605065461},{94371660,396618486},{1308096072,949979895},{1077409785,643586808},
{73924467,1163668217},{150470369,2142214906},{1065875472,1900600059},{296222155,1420331772},
{1565520990,1993526013},{1009252478,588593918},{274726388,513526527},{688388831,1115577088},
{1456469286,1727511297},{1971319136,1242397442},{905967936,1589062403},{145751786,96197380},
{1614279674,1845537541},{211287661,1601510150},{1988620609,1835150087},{279444971,1460357896},
{1644164033,741743369},{1207957248,404560650},{1778381511,2140822283},{2016932117,699546380},
{703068868,155523853},{596638606,976284430},{1276638845,273791759},{1063778324,1231375120},
{963639506,1769667345},{1842344527,1196354322},{2108158125,2007010067},{403176703,1666673428},
{286784989,1281223445},{599260041,1001073430},{233307715,781138711},{1761604321,396479256},
{1154479974,1268730649},{673708795,871807770},{2097672438,2088545051},{646970158,1074957084},
{295697868,1069284125},{925890843,893315870},{1201665805,922815263},{1684534134,1268013856},
{1287648872,1478753057},{583007144,695659298},{1228928729,1572608803},{1219491562,2044271396},
{533724167,467393317},{1398797716,2126023462},{428342479,1359776551},{329252236,1715637032},
{36700090,1467083561},{1259861662,1278331690},{814741999,2052533035},{1757934317,766237484},
{1668281234,793262893},{918550824,476863278},{1855451696,1165912879},{79167337,1371437872},
{460848273,813488945},{59768718,1479084850},{1556083816,572078899},{1729098545,1730276148},
{394263825,1321151285},{1108867005,766069558},{195559051,393063223},{1843917400,1550748472},
{840432061,613641017},{1196422934,1091541818},{557317083,2065865531},{592968598,1241115452},
{1811935880,475982653},{558889942,870255422},{1763701470,376904511},{1599599643,708168512},
{1871704591,1152576321},{1278735998,1834683202},{757070430,336128835},{1944056205,962374468},
{591920023,194517829},{1148188530,649448262},{1107294144,2084113223},{1520432313,827288392},
{1516238006,1570384713},{212860522,1343568714},{2062545074,837147467},{868743559,678812492},
{2064117935,1554504525},{943716603,1554279246},{1381496245,240978767},{1377301955,1909775184},
{1913123264,727309137},{1110439866,1406061394},{1367864783,876936019},{1322251820,581565268},
{2012213533,488942421},{1056962593,2025503574},{1488975081,500812631},{1018689641,337701720},
{1428157790,304147289},{1305998917,578710362},{1289746020,982563675},{1243608766,52632412},
{813693425,1220172637},{159383248,481753950},{1471149323,849140575},{1726477091,1698945888},
{895482196,664742753},{1079506933,803859298},{981989552,1786927971},{429915341,1416293220},
{97517382,1660611429},{409992434,1613896550},{862452115,1732430695},{1414526326,1364585320},
{1446507835,1680481129},{873986430,765246314},{1498936535,1170635627},{989329571,827689836},
{842529213,1432374125},{1432352084,137562990},{1378874811,1195273071},{324533654,1242381168},
{1657795496,311466865},{1099429839,724831090},{1340077573,957160307},{368573761,1677900660},
{639105854,430934901},{14680036,1675742070},{312999340,33753975},{2056253631,1424808824},
{1623716847,1677163385},{2026893545,1966984058},{269483518,792234875},{261619213,1038179196},
{1258813090,319925117},{1559229541,257522558},{1345844729,1020005247},{687340260,1867180928},
{524287001,1177217921},{1074264065,467704706},{2062020876,1494633347},{135266046,222874500},
{1608512516,1379023749},{1278211706,122878854},{432012489,442669959},{127926028,178858888},
{1753215734,618843017},{1788342962,1228049290},{1651504050,1082583947},{400030981,1144085388},
{1860694564,1433045901},{1026029660,444345230},{1256715944,1841712015},{271580666,228744080},
{485489762,1575017361},{52428700,1127689106},{1742205704,615484307},{929560851,355318676},
{1112537016,956595093},{708311737,632138646},{1207432961,2055084951},{1699738456,1533893528},
{98041669,2107468697},{569375682,727419802},{1628959710,1344756635},{916977963,2130160540},
{1885860344,409398173},{78118763,1884363678},{909637945,2087435167},{1882714625,136706976},
{1269298827,90889121},{1612706813,1612712866},{1344271869,598252451},{260046352,1244240804},
{898627918,2003954597},{1874326036,1092651942},{1595929629,1909554087},{1037039687,868809640},
{1778905804,1590332329},{371719483,615648170},{1099954128,1443113899},{1393030561,410553260},
{1184364334,681474989},{160956109,1556589486},{951580905,1716399023},{330825097,3627952},
{1337456137,1446005681},{1774187209,1861073842},{1620571117,1992326067},{1395651996,2064055220},
{1279784567,468372405},{2084565133,1706339254},{449838246,505097143},{1239938756,2035219384},
{188219033,230316985},{1419769197,1302936506},{1171257158,365284283},{608172920,30809020},
{638057279,2069707709},{1919414724,347597758},{1528820893,2114448319},{1727001379,109546432},
{1864364591,1750309825},{81264485,247532482},{789576222,409553859},{10485740,1770228676},
{68681597,1141078981},{990902431,182107078},{1829237349,1917549511},{1319106093,164174792},
{1294464607,1049500617},{224394836,1514363850},{573569978,1253137355},{434633923,178641868},
{307232182,1367407565},{340262264,1308240846},{300416451,594869199},{1292891742,372263888},
{2144334049,687213521},{869267846,43084754},{1914696124,463535059},{694680275,1881975764},
{1593308195,948857813},{118488862,1975901142},{2133323886,614230999},{52952987,411380696},
{1540879494,1692789721},{110100270,1758313434},{91750225,1067375579},{445119663,1669270492},
{1192228638,1576446941},{1145567097,2042379230},{937425156,443743199},{1951920513,1451609056},
{783284779,2069105633},{870316420,121220066},{1128789913,912030691},{1616901113,116992996},
{1564472419,441760741},{1840247376,478665702},{9437166,154205159},{1138751364,1447844840},
{1722807087,1374886889},{1348466166,1374718954},{1113585588,2109402091},{290454998,1532697580},
{530578444,1012304877},{747633262,1630448622},{1152382827,868236271},{498072650,646663152},
{1588589611,890379249},{303037886,779246578},{237502011,1197628403},{1872228882,709254132},
{1224210145,2087652341},{539491323,543210486},{659028759,568798199},{85983068,1572178936},
{1504179403,62708729},{977270968,1527004154},{1387263407,1763699707},{1050146861,232107004},
{713554608,1390672893},{713030320,1768889342},{1217918703,50478079},{736623235,818727936},
{1392506273,1505782785},{1496315098,933526530},{517471269,999648259},{774896186,436845572},
{480246892,1436326917},{523238426,647216134},{284163554,1738464263},{2022699249,244255752},
{1442313537,697375753},{1657271208,2120375306},{1957687660,1853586443},{116916001,306162700},
{166723266,1986026509},{820509156,1145302030},{1371010505,1511394319},{1906831831,553311248},
{2115498048,2024590353},{1738535695,576445458},{2066739391,267009043},{1995960637,579337236},
{1180694324,1541106709},{876607866,2127031318},{283639267,1819311127},{761264724,847297560},
{77070189,973720601},{1781527226,464337946},{1530918045,578800667},{1723855665,341077020},
{1701835604,976698397},{683670248,1607244830},{1874850313,489339935},{32505794,1809169440},
{1178072889,535407649},{509606964,154844194},{1011873910,2003618851},{1114109876,932019236},
{901773640,457149477},{847772079,243866662},{154140378,1791671335},{54001561,594922536},
{2056777932,1158339625},{1670378383,243960874},{1123022754,1091669035},{913832241,1270262828},
{1572336714,1770732589},{2075652242,1027763246},{1282406002,34241583},{2117070918,1294371888},
{1841295949,1020103729},{278396397,1912523826},{1890578927,1892854835},{670038786,230853684},
{802683398,1099455541},{603978624,1740782646},{394788111,1961794615},{510655538,873348152},
{1636824032,540351545},{383253797,786820154},{718273190,228777019},{1254094506,895872060},
{1105721289,1097804861},{1012398197,1361071166},{1968697688,1783442495},{213909096,1702562880},
{1884287494,811584577},{1276114558,170400834},{2046816464,1402129475},{512752687,520219716},
{1304950343,893594693},{1372059080,975301702},{672135935,883436615},{1912598980,888605768},
{320863644,860675145},{857209245,712825930},{1072691202,2105175115},{12582888,2128145484},
{1221064424,2064022605},{1570239566,215002190},{892336476,2064890959},{1487926508,1246248016},
{992475291,642325585},{1727525666,328019026},{987756708,29600851},{2099245151,869686356},
{697301710,425426005},{1629483997,417897558},{177209006,1157446743},{438828219,1408060504},
{903870788,699436121},{1646261180,1537616986},{1050671151,1482497115},{805304832,1904942172},
{35127229,1733491805},{843577783,146111582},{951056621,334052447},{339213689,2091494496},
{263716361,784194657},{1642066884,916175970},{2048389322,1008082019},{756021854,91155556},
{1853878832,1979477093},{1581773882,415083622},{1997009203,1098939495},{88080216,109354088},
{282066406,233995369},{250084899,1404902506},{207093365,627141739},{1171781451,1478597740},
{373292344,880098413},{402652416,1757432942},{1315436085,1951292527},{872937857,1734085744},
{1955066225,1825377393},{153091804,604454002},{396885261,655338611},{1229977304,1796684916},
{268434944,1009159285},{1209005822,2032479350},{1870131747,802475127},{41942960,1259236472},
{1411904892,1423998073},{1135605642,758172794},{2094526574,1278921851},{1750070011,651754620},
{590347162,191782013},{607648634,1949355134},{552598499,1271676031},{1301280337,227458176},
{2045243598,95947905},{538967036,221932674},{1469052174,517131395},{1882190340,1931525252},
{1655698357,1190071429},{1736962832,1100926086},{1984426306,888818823},{1027602520,675851400},
{1948250494,1461664905},{23068628,1708555402},{742914679,335293579},{1613755398,25832588},
{1947726205,430853261},{1835004502,1614929038},{1670902670,1200778383},{566754247,1324928144},
{1407710595,1405344913},{191889042,2128567442},{1591735340,2041212051},{2099769465,1520098452},
{2076700850,2028391573},{1936716189,1317047446},{1529345180,1753767063},{367000900,2046479512},
{1437594954,134044825},{1084749804,1339391130},{1191180064,790989979},{1672475530,17878172},
{178257580,393931933},{1653076911,1407261854},{1127217050,1225555103},{770177603,158256288},
{990378144,711527585},{540539897,667380898},{1817703029,999480483},{1322776105,962481316},
{1886384630,20855973},{1315960371,914996390},{451411107,977247399},{856684959,1546206376},
{1515189430,388770985},{856160671,544799914},{58720144,1564286123},{2035282161,1375685804},
{1790964393,691440813},{1037563973,1309883566},{1643115461,1080564911},{1686631279,1417465008},
{8912879,360627377},{547879915,558443698},{1621095410,1782217907},{193986190,1375968436},
{1033369684,1391475893},{1810887300,1378118838},{1648358328,174820535},{1008728188,185711800},
{959969497,1853971641},{1549792374,1771383994},{106430263,253406395},{818936294,497913020},
{1236268746,2009717949},{165150405,1647492286},{795343379,1051221183},{484441189,1765301440},
{2057302190,1640832193},{536869888,1947229378},{1329067547,534236355},{1388311978,249801924},
{527432722,302148805},{1898443238,1882500294},{2010640674,753581255},{1968173406,164895944},
{1272444549,1472818377},{1649406904,1121094858},{1087895529,296320203},{1412429178,333065420},
{1104672709,1268858061},{1042282557,414391502},{799013388,1436306639},{1098905553,370089168},
{96468808,1998982353},{737671810,1633328338},{1900016088,280206547},{1499985109,2069630164},
{1823470195,331807957},{2100293765,1776975062},{1218967276,1854160087},{221773401,1142365400},
{1412953465,1242631385},{1161819994,963247322},{902297929,1943387355},{1067972619,19381468},
{2002252062,1887857885},{1400894869,181259486},{1563423838,94522591},{879229299,2106354912},
{1351087601,1268309217},{2124410945,347966690},{1685058427,1626152163},{412089582,1251634404},
{1960309102,1556737253},{24641489,866614502},{893385048,855882983},{947910896,515206376},
{266862083,999406825},{398458120,971562218},{1935143321,130030827},{549977064,1146891500},
{1290794594,1624526061},{1740108557,1528458478},{2103963754,184155375},{1380447671,939498736},
{441449654,1355521265},{1540355214,1516997874},{1328543260,68754675},{84410207,2105408756},
{1865937443,1693613301},{1900540379,157859062},{1672999817,962497783},{1380971958,1781714168},
{305659321,363773177},{1851257397,1110195450},{301465025,616586491},{1669854110,622050556},
{284687842,1341869309},{1601172504,1859325182},{1869083165,1151483135},{950008044,984767744},
{1876947461,2025278721},{974125246,1377877250},{1038088262,1822354691},{48758691,197627140},
{279969258,1424260357},{1766322903,1692851462},{114818853,550538503},{200801921,1603566856},
{576191413,599620873},{232259141,1090800906},{694155988,1908284683},{471334013,1436986636},
{1167587150,1587657997},{381680936,312343822},{1382020534,1289886991},{239599159,1162935568},
{552074211,363113745},{154664665,212327698},{980416690,1099885843},{315620774,351120660},
{2102915237,1851243797},{1280308856,343768342},{1167062865,248937751},{656407324,351915288},
{1262483098,369687833},{177733293,313494810},{773323325,2103921947},{1828713072,1661160732},
{977795255,122834205},{1443362114,1075272990},{2146431944,1582796063},{1586492467,471100704},
{731904652,597122337},{39321525,2142412066},{1578628166,2123435299},{1179645750,1267911972},
{1244133055,1675210021},{332922245,2107997478},{974649534,263429415},{1368913359,391576872},
{1235220172,1433046313},{468188291,827059498},{315096487,355679531},{1606939659,320478508},
{1908928970,773872941},{1661989790,1770179886},{20447193,333569327},{635960133,1511025968},
{1979183470,1925692721},{860354967,958332210},{537918462,529644851},{271056380,558308660},
{1740632848,1771478325},{340786550,717385014},{941095168,1206672695},{1204287240,1054170424},
{2104488064,1287851321},{787479076,448367930},{953153768,144432443},{458226838,1247141180},
{1413477753,1691102525},{429391056,1694936382},{304610747,559668543},{225967697,959130944},
{2025320693,1922952513},{410516721,2136583490},{1040709695,713088323},{1058011167,1456483652},
{1359476193,2081090885},{1502606542,1008684358},{568851395,596626759},{825227738,1052114248},
{1435497807,497556809},{1710748482,1924345162},{1341650433,1396538699},{1637348301,1544240460},
{1156577124,1742462285},{1592259620,874397006},{391118102,631393615},{782760491,1642155344},
{1527248032,1219104081},{804780547,603774290},{536345603,429268307},{1055914023,1432624468},
{1570763853,243526997},{215481957,338980182},{660077333,1526406487},{1212151544,1000275288},
{854587811,1529584985},{944240887,812485978},{1561326687,2000014683},{642251575,1487367516},
{214957671,971570525},{199753347,1045171550},{100138817,808918367},{2049962210,1041657184},
{206044791,996343137},{1600648213,485936482},{2013786374,1541975395},{492829780,962735460},
{274202101,355478885},{1777857218,1941323110},{1318581805,1251851623},{1978134853,1251315048},
{1742729994,231611753},{1373631944,1646251370},{1147139957,1491635563},{899152205,1785490796},
{1785197239,677477741},{578812848,2071461230},{1658844072,305659247},{1907356122,73809264},
{1266153105,1685458289},{612367217,554237298},{1286076011,1998814579},{1682961275,246279540},
{1911026133,133102965},{2006446355,1978592630},{1999630618,1958485367},{2143285567,1184210296},
{404749564,776015225},{655358750,2065423738},{443022515,1873132923},{887093604,2137349500},
{1994387755,1639562621},{421002461,1434185086},{520616991,1779191167},{232783428,1639464320},
{871889281,1953348993},{986708134,1460231554},{1465906453,739532163},{2072506525,931216772},
{1542976641,1868447109},{437255359,1878515078},{1693447013,988786055},{134217472,779685256},
{76021615,723021193},{1026553950,1856417162},{22544341,13032843},{1548219511,1605537164},
{6815731,420257165},{1572861001,525217166},{1708651342,1674075535},{921696546,92470672},
{1898967514,1015487889},{1846538829,978087314},{837810626,1181416851},{1660941219,249965972},
{211811948,1188711829},{1408759169,1625705878},{1041233984,1862143383},{1820324468,1264876952},
{1034942539,439156121},{231210567,528907674},{50855839,1145818523},{1336407565,640744860},
{668465926,738024861},{2100818145,751570334},{2081419409,224292255},{73400180,200510880},
{727710356,1525497249},{979368117,1243385250},{459275412,1910132131},{1895297508,37424548},
{1563948122,1457175973},{1309144642,1278463398},{796916241,1193930151},{681048813,1927601576},
{2015359238,264146345},{475528309,440663466},{1988096305,548634027},{440401080,656096684},
{335543680,390880685},{635435844,526708142},{1257240226,123960751},{785381926,671931824},
{81788772,351583665},{859306394,1520889266},{1689777005,1206107571},{1955590511,1441906100},
{348650855,336612789},{1979707715,16403894},{470809726,901479863},{1051195437,1329638840},
{823130590,1657413049},{1620046835,853167546},{2027417840,1561849275},{70254458,1304624572},
{2098196708,464559549},{785906213,1255013822},{2008019227,1728847295},{711457459,1394093504},
{2113400928,569789889},{984610986,1969884610},{1632629720,949104067},{788003363,267836868},
{560462803,1243205061},{192937616,194579910},{1168635727,139820487},{1732244254,1840319944},
{986183848,1487879625},{858782106,535940554},{1028126814,1845714379},{1448080697,1794538956},
{1590686758,2130038221},{1351611888,1962147278},{1794634404,1902640591},{1536160910,1799523792},
{1524102309,1382624721},{1477965056,1933954514},{985135273,1605983699},{270007805,1982094804},
{513276973,604208597},{917502250,1265466838},{1098381265,1240075735},{1288697446,1538649560},
{676330230,1424023001},{1080555511,1762557402},{2091380874,385510875},{1806693005,1415392732},
{107478835,1212169693},{1961357690,724012510},{1760031465,1516309983},{1982853450,1154801120},
{1862267426,766504417},{1191704351,871210466},{834140618,2048855523},{346029420,1158692324},
{111148845,1486667237},{2145908329,1742003686},{2101342337,997965287},{685243109,1584639464},
{1076885498,2074398185},{409468148,1334795754},{1134032782,1615375851},{1616376822,1638518252},
{1933570467,2033503725},{341835124,857800174},{1107818432,464518639},{1622143978,772787696},
{686815970,1637965297},{674233082,1176141298},{2126508094,1895489011},{146276073,456207860},
{1496839386,1373613557},{529529870,2092686838},{1523578025,2091818487},{407895286,1002524152},
{1005058179,739765753},{505412668,1211862522},{1019738215,1643998715},{1823994499,1887383036},
{1306523204,1617219069},{733477516,1212112382},{983562412,1025588735},{8388592,1204444672},
{1830285922,1560772097},{2125983870,1681399298},{1831334514,1573076483},{704641728,213323268},
{1655174062,2007059973},{1649931192,1321061894},{2070933661,1054129671},{749730412,1663016456},
{1960833381,1339063817},{1294988892,1985695242},{226491984,1248050699},{1524626598,922635788},
{559938516,1849908749},{129498889,1102560782},{393215250,977833487},{1987047741,729288208},
{293076433,1044639249},{1884811776,1247784466},{884472171,820280851},{1659368368,1526853140},
{299892165,594386453},{553122786,1528327702},{1250948782,1068240407},{994048154,505732632},
{1087371240,722251289},{547355628,668405274},{1439692104,1777679899},{185073311,1120202268},
{580909996,1711762973},{1607463942,1978596894},{312475052,1275923999},{1464333591,2081603104},
{1182267185,46214689},{2111303799,1724096034},{691534553,1866579491},{453508255,1036029476},
{1013446775,860810789},{1517810869,73293350},{957872349,245792295},{1472722183,212225576},
{355990873,1670315561},{634911557,1131232810},{93847373,1316314667},{502266946,351723052},
{2017980666,645635629},{991426719,1189092910},{82837346,1132224047},{1083701229,741060144},
{531102731,1978097201},{523762713,1454235186},{1291318881,384572979},{1675096970,295611956},
{213384810,1910394421},{1743254278,1136541238},{656931611,401489463},{1706029899,732065336},
{371195196,302816825},{1588065327,1103896122},{826800599,1276100155},{1054865444,399736380},
{1202714378,1152941629},{1073215489,807833150},{172490423,1837588031},{1869607445,1558085184},
{1681388420,89304641},{828897748,1006812738},{714078894,966086211},{1197995797,1107783236},
{1783624375,719523397},{1095235543,330989126},{1990717746,209346119},{529005583,522526280},
{122683158,2047663689},{1405089160,1725193802},{797440527,1977155147},{1484256500,1706901068},
{580385709,404557389},{108003122,959618638},{627047252,2039463503},{1916793280,126197328},
{1281881715,828898897},{567278534,337935954},{927463703,1699397203},{2094002281,716025428},
{1016592493,1140424277},{1280833142,61636182},{46137256,676585047},{928512278,262131288},
{644873010,2046635609},{1242560190,490180186},{1136129929,447565403},{958920923,1635364444},
{1999106336,2099175005},{30408646,26693214},{1697641307,11865695},{1791488679,1769815648},
{1902113246,1942789729},{693631701,1071951458},{1887957500,1073749603},{1969746266,1483841124},
{259522066,1097703013},{1888481789,2016956006},{683145961,1336970855},{1268774542,1840610920},
{1926230456,474050153},{1510995134,55844458},{790100509,233082475},{1958211961,1262669420},
{1846014532,624000621},{1541403780,884477550},{45088682,1925086831},{1546122369,2108231280},
{1233647311,214724209},{13631462,890580594},{926939416,230157939},{1963979107,1014333044},
{1479537914,165760629},{1389884838,1277247094},{1399846290,1206763127},{677378804,1468104312},
{1634202584,1050975865},{1181742900,372280954},{1624241127,924577403},{1205860102,1667370620},
{1139799940,930328189},{414711017,1637527166},{1425012070,312262271},{1614803962,510992000},
{915929389,135544449},{545258480,1111215746},{1035466825,1387695747},{2140139565,449162884},
{698350284,54197893},{899676493,992185990},{826276312,1479110279},{705690303,938438280},
{1587541040,1464712841},{67633023,186695306},{1534063765,1352822411},{1630532573,1115405964},
{924317981,205872781},{1401943438,1166175886},{493354068,473992847},{488635484,809504400},
{353369438,1381371537},{1175975743,1020329618},{490732632,1052745363},{645397297,829902484},
{1574433863,60939925},{1668805522,1835732630},{1711797059,1559752343},{1752167155,967794328},
{426769619,403742361},{351272290,2047434394},{365952326,1089093275},{1070594056,434966172},
{482868327,1148276381},{176684719,122797726},{1457517862,1639964319},{236977724,478199456},
{235404863,728661665},{628095826,1320308386},{55050135,299650723},{1273493124,1555644068},
{2001727799,274165413},{1027078237,1329651366},{909113659,758259367},{1974464848,1550405288},
{882375021,2078723753},{987232423,1116180138},{142606064,1918406315},{1569190997,1169817260},
{108527409,1573498541},{386399519,878550702},{1686106994,1942392495},{2077225130,363630256},
{866122125,1707908785},{1844965974,1134050994},{866646411,764698291},{1551889525,153452212},
{1256191653,1118826165},{1841820249,132304566},{1436022099,1982217911},{75497328,811491000},
{422051036,23826105},{1661465503,1043955386},{510131251,114880187},{931133712,53759676},
{611318642,1269231293},{188743320,1162370750},{786430500,500076223},{304086460,1507176128},
{457178264,1990590145},{1021311076,2037767874},{1765274330,590610115},{1578103870,1263226564},
{282590693,289361605},{1741157129,74567366},{549452776,1024667335},{1951396224,1378909896},
{181927589,2033925833},{454032542,176824010},{210239087,368094923},{1878520329,1252216524},
{270532092,90934989},{2127032563,364240590},{466615430,1715809999},{1964503406,660430544},
{1299183186,412081873},{501742659,1527889618},{1203762952,944078547},{1311766075,1189043924},
{496499790,220446421},{37748664,677404374},{2049437897,1805885143},{641203001,513441496},
{1767371492,1644736217},{1352660462,865713882},{1124595616,1391574747},{997193874,741756636},
{725613210,2028486365},{1893724644,232795870},{729807506,1727958751},{1815605892,325951200},
{932706574,1739353825},{208666226,1092644578},{944765177,1297948387},{540015610,1661914852},
{144178925,637320933},{1939337613,661110502},{806353406,964374247},{1552413810,290864872},
{1790440107,832401129},{1016068208,848654058},{577239987,207097579},{1257764513,183615212},
{1069545480,1776901869},{864549263,52485870},{503315520,726241007},{763361872,1476779760},
{1755837163,1258409713},{1858073141,1946918642},{1508897986,2080038643},{638581566,16469748},
{1966600538,1293856501},{1544549506,1228082934},{967833804,1490583287},{162004683,1502437112},
{1024456799,885186297},{1784148664,163081978},{483392615,1695502075},{187170459,1542999804},
{619182947,547954429},{174587571,1938013950},{2084040835,1091297023},{1029699670,10858240},
{617085799,339619585},{1861743138,1142615810},{1082128370,1278856963},{1458042148,1647574788},
{1272968838,390291205},{392166677,28262150},{2004873490,983498503},{1354233321,1502080776},
{261094926,836779785},{508034103,1767587594},{1101526987,1504857867},{769129029,1686515468},
{1946677632,993414925},{514849834,69390},{1879568903,1188658959},{1895821801,83341072},
{2047865053,1363713809},{102235965,2139287314},{575142840,1120812819},{1438119243,1218117396},
{1700262754,165138197},{958396638,322969366},{1481110775,1532686103},{1654649796,1393245976},
{1483732210,1569099545},{789051936,1620877082},{1826615910,2117287707},{1820848752,1087737628},
{1486877932,433037085},{1160771418,852061982},{360185169,138272543},{765983307,245092128},
{1526199457,1628565281},{1209530109,1817689890},{405273851,1963613987},{607124346,1396080420},
{1924133292,756698917},{966785228,845786918},{1683485557,1671270183},{347077994,240623400},
{2000154906,1608417065},{2114973776,1760411434},{779090482,1010204459},{362806604,101244716},
{1849160254,1326059309},{344980846,215838510},{726661782,2014416687},{1995436338,1693617968},
{1644688328,1795039025},{514325549,13872946},{1909977541,335400755},{221249114,1192464180},
{961542359,341327669},{1226307296,98959158},{1006106754,957894455},{253754908,1572937528},
{1767895765,1805020985},{2090856646,388575034},{844102071,1040600891},{349699430,535424828},
{955775206,618700605},{2011164933,798424894},{2028466476,10493759},{368049474,1374191424},
{1771565780,413110081},{1117255601,198745922},{1938289081,1351003971},{2055205058,1668042564},
{956299488,1984323397},{2077749423,1131831110},{1242035903,333782855},{277872110,163655496},
{640154427,1211813705},{629144401,1236131658},{85458781,876109643},{1848111675,1728659276},
{546831341,773615437},{1535112342,631439182},{1494742237,188374863},{42991534,1510838096},
{541588471,311209809},{1667232663,452153170},{1487402219,1739108179},{1297610332,562634580},
{1915744700,632745813},{1251473069,1999081302},{121634584,1260056407},{877132151,2066693976},
{2032136430,389779289},{1045952565,1599995738},{1342699007,588607323},{1577055303,870367068},
{1531442332,1811492701},{47185830,1022861150},{1803022993,1953668959},{1152907114,511618912},
{521665566,1015238497},{922220833,1281830754},{1574958151,95248227},{811596278,1323573092},
{1777332931,1999961957},{1598551068,566222694},{1501557970,811560807},{850917801,1768394600},
{830470608,1060622185},{930085138,533872490},{65011588,844611435},{1153431400,374513516},
{29360072,1494298477},{942668026,299462510},{1423439207,1167298415},{834664904,260960112},
{387972380,280002417},{260570639,343703410},{1066399758,1386008435},{1214248693,624566132},
{1169160012,187506549},{1055389734,818343798},{2060447948,573767543},{957348062,758202232},
{1969221987,1430163321},{525859861,242364282},{1342174721,1490481019},{346553707,107470716},
{730331791,446443389},{1783100087,1413513086},{42467247,909217663},{1271920262,1981329280},
{1137702790,1842220929},{1838674509,1721323394},{1560278112,1180065667},{1511519421,1592819588},
{786954789,449359749},{1915220419,746999686},{2097148,419950471},{1229453016,375807880},
{1188558632,1330823049},{1369961935,533278602},{170393275,1388908427},{2073030837,576724876},
{720370338,130342797},{941619452,1134858126},{119013149,1964539791},{544734193,601821072},
{364903752,1583996817},{553647072,462266258},{970455239,470339475},{783809065,153440148},
{524287,596864917},{701496006,98529174},{117964575,996954007},{435158210,972029848},
{1479013630,1540116377},{2085613709,1077682074},{1764750047,1938231195},{2120216665,995553180},
{1103099852,1690484637},{1143994234,850481054},{265289222,33873823},{835713478,728088480},
{978843831,1525624737},{1581249593,441778082},{2080895147,1874853795},{1426584933,369991588},
{254279195,2099724197},{460323986,1409290150},{1285551726,1021218727},{1861218852,2112212904},
{1696592734,494608297},{1350039025,1826570154},{1800925847,1927450539},{708836025,535064492},
{133168898,58904493},{840956348,472944558},{521141278,153726895},{325582227,250347440},
{1375204801,1840922545},{15204323,472407986},{1542452355,567439283},{105381687,1519947700},
{171441849,2006716341},{1058535454,754433974},{2009067794,1395343287},{1799877271,265387960},
{2112352326,1109499833},{1733292822,540733370},{964163793,1384423355},{569899969,826998716},
{792197657,253853629},{389020954,617639870},{577764274,320331711},{590871449,927621056},
{832567758,653774785},{1293416030,199282626},{146800360,1832341443},{83361633,1224208324},
{1664086942,1219624901},{1718612787,1817325510},{1089992675,1091321799},{669514499,1681145800},
{1062205462,1437638601},{1554510958,217165770},{69205884,750239691},{583531431,1748426700},
{436206785,568950733},{428866766,1871622094},{1763177187,514392015},{148897508,173191120},
{1255667367,779665361},{888666466,681385938},{2074603700,1935437779},{59244431,890957780},
{2060972230,356339669},{1827140197,1019531222},{21495767,540565463},{1648882629,214192088},
{101187391,445226969},{900200781,877297626},{1356330471,1482711003},{598735754,838938588},
{1200092946,290467805},{1682436991,24592350},{1688728428,61296607},{1456993577,173354976},
{684194536,1000185825},{306707895,356417506},{1566045271,1299181539},{1091565534,579026916},
{907016511,458018789},{1183840047,1023938534},{1980232008,1544904679},{1738011416,1008762856},
{2021126393,563601385},{387448093,339034090},{189791894,1855213547},{1562375264,234094572},
{2014834944,1585672173},{1160247139,1175506926},{1887433218,1318850543},{588250017,449249264},
{611842929,1815785457},{696253139,1936674802},{74973041,1390813171},{1754788594,1673334772},
{956823776,167489525},{1173354306,2111897590},{652737316,796594167},{780139056,1157468152},
{1035991112,557969401},{1591211048,907370490},{1836053075,1329336315},{239074872,365875196},
{1092089821,1687924733},{359660883,1123516414},{1702359889,745918463},
};

